# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
Functions for modifying the xml input file of Fleur with explicit xpath arguments
These can still use the schema dict for finding information about the xpath
"""
from lxml import etree
from masci_tools.util.xml.common_functions import eval_xpath

######################CREATING/DELETING TAGS###############################################


def xml_create_tag_schema_dict(xmltree,
                               schema_dict,
                               xpath,
                               base_xpath,
                               element,
                               create_parents=False,
                               occurrences=None):
    """
    This method evaluates an xpath expression and creates a tag in a xmltree under the
    returned nodes.
    If there are no nodes evaluated the subtags can be created with `create_parents=True`

    The tag is always inserted in the correct place if a order is enforced by the schema

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param xpath: a path where to place a new tag
    :param base_xpath: path where to place a new tag without complex syntax ([] conditions and so on)
    :param element: a tag name or etree Element to be created
    :param create_parents: bool optional (default False), if True and the given xpath has no results the
                           the parent tags are created recursively
    :param occurrences: int or list of int. Which occurence of the parent nodes to create a tag.
                        By default all nodes are used.

    :raises ValueError: If the nodes are missing and `create_parents=False`

    :returns: xmltree with created tags
    """
    from masci_tools.util.xml.xml_setters_basic import xml_create_tag
    from masci_tools.util.xml.common_functions import check_complex_xpath, split_off_tag

    check_complex_xpath(xmltree, base_xpath, xpath)

    tag_info = schema_dict['tag_info'][base_xpath]

    if not etree.iselement(element):
        #Get original case of the tag
        element_name = (tag_info['simple'] | tag_info['complex']).original_case[element]
        try:
            element = etree.Element(element_name)
        except ValueError as exc:
            raise ValueError(f"Failed to construct etree Element from '{element_name}'") from exc
    else:
        element_name = element.tag

    if len(tag_info['order']) == 0:
        tag_order = None
    else:
        tag_order = tag_info['order']

    several_tags = element_name in tag_info['several']

    parent_nodes = eval_xpath(xmltree, xpath, list_return=True)

    if len(parent_nodes) == 0:
        if create_parents:
            parent_xpath, parent_name = split_off_tag(base_xpath)
            complex_parent_xpath, _ = split_off_tag(xpath)
            xmltree = xml_create_tag_schema_dict(xmltree,
                                                 schema_dict,
                                                 complex_parent_xpath,
                                                 parent_xpath,
                                                 parent_name,
                                                 create_parents=create_parents)
        else:
            raise ValueError(f"Could not create tag '{element_name}' because atleast one subtag is missing. "
                             'Use create=True to create the subtags')

    return xml_create_tag(xmltree, xpath, element, tag_order=tag_order, occurrences=occurrences, several=several_tags)


def eval_xpath_create(xmltree,
                      schema_dict,
                      xpath,
                      base_xpath,
                      create_parents=False,
                      occurrences=None,
                      list_return=False):
    """
    Evaluates and xpath and creates tag if the result is empty

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param xpath: a path where to place a new tag
    :param base_xpath: path where to place a new tag without complex syntax ([] conditions and so on)
    :param create_parents: bool optional (default False), if True also the parents of the tag are created
                           if they are missing
    :param occurrences: int or list of int. Which occurence of the parent nodes to create a tag if the tag is missing.
                        By default all nodes are used.
    :param list_return: if True, the returned quantity is always a list even if only one element is in it

    :returns: list of nodes from the result of the xpath expression
    """
    from masci_tools.util.xml.common_functions import check_complex_xpath, split_off_tag

    check_complex_xpath(xmltree, base_xpath, xpath)

    nodes = eval_xpath(xmltree, xpath, list_return=True)

    if len(nodes) == 0:
        parent_xpath, tag_name = split_off_tag(base_xpath)
        complex_parent_xpath, _ = split_off_tag(xpath)
        xmltree = xml_create_tag_schema_dict(xmltree,
                                             schema_dict,
                                             complex_parent_xpath,
                                             parent_xpath,
                                             tag_name,
                                             create_parents=create_parents,
                                             occurrences=occurrences)
        nodes = eval_xpath(xmltree, xpath, list_return=True)

    if len(nodes) == 1 and not list_return:
        nodes = nodes[0]

    return nodes


def xml_set_attrib_value(xmltree,
                         schema_dict,
                         xpath,
                         base_xpath,
                         attributename,
                         attribv,
                         occurrences=None,
                         create=False):
    """
    Sets an attribute in a xmltree to a given value. By default the attribute will be set
    on all nodes returned for the specified xpath.
    If there are no nodes under the specified xpath a tag can be created with `create=True`.
    The attribute values are converted automatically according to the types of the attribute
    with :py:func:`~masci_tools.util.xml.converters.convert_to_xml()` if they
    are not `str` already.

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param xpath: a path where to set the attributes
    :param base_xpath: path where to place a new tag without complex syntax ([] conditions and so on)
    :param attributename: the attribute name to set
    :param attribv: value or list of values to set
    :param occurrences: int or list of int. Which occurence of the node to set. By default all are set.
    :param create: bool optional (default False), if True the tag is created if is missing

    :raises ValueError: If the conversion to string failed
    :raises ValueError: If the tag is missing and `create=False`
    :raises ValueError: If the attributename is not allowed on the base_xpath

    :returns: xmltree with set attribute
    """

    from masci_tools.util.xml.xml_setters_basic import xml_set_attrib_value_no_create
    from masci_tools.util.xml.converters import convert_to_xml
    from masci_tools.util.xml.common_functions import check_complex_xpath, split_off_tag

    check_complex_xpath(xmltree, base_xpath, xpath)

    if create:
        nodes = eval_xpath_create(xmltree,
                                  schema_dict,
                                  xpath,
                                  base_xpath,
                                  create_parents=True,
                                  occurrences=occurrences,
                                  list_return=True)
    else:
        nodes = eval_xpath(xmltree, xpath, list_return=True)

    if len(nodes) == 0:
        raise ValueError(f"Could not set attribute '{attributename}' on path '{xpath}' "
                         'because atleast one subtag is missing. '
                         'Use create=True to create the subtags')

    _, tag_name = split_off_tag(base_xpath)

    attribs = schema_dict['tag_info'][base_xpath]['attribs']
    if attributename not in attribs:
        raise ValueError(
            f"The key '{attributename}' is not expected for this version of the input for the '{tag_name}' tag. "
            f'Allowed attributes are: {attribs.original_case.values()}')
    attributename = attribs.original_case[attributename]

    converted_attribv, _ = convert_to_xml(attribv, schema_dict, attributename, text=False)

    return xml_set_attrib_value_no_create(xmltree, xpath, attributename, converted_attribv, occurrences=occurrences)


def xml_set_first_attrib_value(xmltree, schema_dict, xpath, base_xpath, attributename, attribv, create=False):
    """
    Sets the first occurrence attribute in a xmltree to a given value.
    If there are no nodes under the specified xpath a tag can be created with `create=True`.
    The attribute values are converted automatically according to the types of the attribute
    with :py:func:`~masci_tools.util.xml.converters.convert_to_xml()` if they
    are not `str` already.

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param xpath: a path where to set the attribute
    :param base_xpath: path where to place a new tag without complex syntax ([] conditions and so on)
    :param attributename: the attribute name to set
    :param attribv: value or list of values to set
    :param create: bool optional (default False), if True the tag is created if is missing

    :raises ValueError: If the conversion to string failed
    :raises ValueError: If the tag is missing and `create=False`
    :raises ValueError: If the attributename is not allowed on the base_xpath

    :returns: xmltree with set attribute
    """

    return xml_set_attrib_value(xmltree,
                                schema_dict,
                                xpath,
                                base_xpath,
                                attributename,
                                attribv,
                                create=create,
                                occurrences=0)


def xml_set_text(xmltree, schema_dict, xpath, base_xpath, text, occurrences=None, create=False):
    """
    Sets the text on tags in a xmltree to a given value. By default the text will be set
    on all nodes returned for the specified xpath.
    If there are no nodes under the specified xpath a tag can be created with `create=True`.
    The text values are converted automatically according to the types
    with :py:func:`~masci_tools.util.xml.converters.convert_to_xml()` if they
    are not `str` already.

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param xpath: a path where to set the text
    :param base_xpath: path where to place a new tag without complex syntax ([] conditions and so on)
    :param text: value or list of values to set
    :param occurrences: int or list of int. Which occurence of the node to set. By default all are set.
    :param create: bool optional (default False), if True the tag is created if is missing

    :raises ValueError: If the conversion to string failed
    :raises ValueError: If the tag is missing and `create=False`

    :returns: xmltree with set text
    """
    from masci_tools.util.xml.xml_setters_basic import xml_set_text_no_create
    from masci_tools.util.xml.converters import convert_to_xml
    from masci_tools.util.xml.common_functions import check_complex_xpath, split_off_tag

    check_complex_xpath(xmltree, base_xpath, xpath)

    if create:
        nodes = eval_xpath_create(xmltree,
                                  schema_dict,
                                  xpath,
                                  base_xpath,
                                  create_parents=True,
                                  occurrences=occurrences,
                                  list_return=True)
    else:
        nodes = eval_xpath(xmltree, xpath, list_return=True)

    if len(nodes) == 0:
        raise ValueError(f"Could not set text on path '{xpath}' because atleast one subtag is missing. "
                         'Use create=True to create the subtags')

    _, tag_name = split_off_tag(base_xpath)

    converted_text, _ = convert_to_xml(text, schema_dict, tag_name, text=True)

    return xml_set_text_no_create(xmltree, xpath, converted_text, occurrences=occurrences)


def xml_set_first_text(xmltree, schema_dict, xpath, base_xpath, text, create=False):
    """
    Sets the text on the first occurrence of a tag in a xmltree to a given value.
    If there are no nodes under the specified xpath a tag can be created with `create=True`.
    The text values are converted automatically according to the types
    with :py:func:`~masci_tools.util.xml.converters.convert_to_xml()` if they
    are not `str` already.

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param xpath: a path where to set the text
    :param base_xpath: path where to place a new tag without complex syntax ([] conditions and so on)
    :param text: value or list of values to set
    :param create: bool optional (default False), if True the tag is created if is missing

    :raises ValueError: If the conversion to string failed
    :raises ValueError: If the tag is missing and `create=False`

    :returns: xmltree with set text
    """
    return xml_set_text(xmltree, schema_dict, xpath, base_xpath, text, create=create, occurrences=0)


def xml_add_number_to_attrib(xmltree,
                             schema_dict,
                             xpath,
                             base_xpath,
                             attributename,
                             add_number,
                             mode='abs',
                             occurrences=None):
    """
    Adds a given number to the attribute value in a xmltree. By default the attribute will be shifted
    on all nodes returned for the specified xpath.
    If there are no nodes under the specified xpath an error is raised

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param xpath: a path where to set the attributes
    :param base_xpath: path where to place a new tag without complex syntax ([] conditions and so on)
    :param attributename: the attribute name to change
    :param add_number: number to add/multiply with the old attribute value
    :param mode: str (either `rel` or `abs`).
                 `rel` multiplies the old value with `add_number`
                 `abs` adds the old value and `add_number`
    :param occurrences: int or list of int. Which occurence of the node to set. By default all are set.

    :raises ValueError: If the attribute is unknown or cannot be float or int
    :raises ValueError: If the evaluation of the old values failed
    :raises ValueError: If a float result is written to a integer attribute

    :returns: xmltree with shifted attribute
    """
    from masci_tools.util.schema_dict_util import read_constants
    from masci_tools.util.xml.converters import convert_from_xml
    from masci_tools.io.common_functions import is_sequence
    from masci_tools.util.xml.common_functions import check_complex_xpath, split_off_attrib, split_off_tag

    check_complex_xpath(xmltree, base_xpath, xpath)

    if attributename not in schema_dict['attrib_types']:
        raise ValueError(
            f"You try to shift the attribute:'{attributename}' , but the key is unknown to the fleur plug-in")

    possible_types = schema_dict['attrib_types'][attributename]

    if not etree.iselement(xmltree):
        constants = read_constants(xmltree.getroot(), schema_dict)
    else:
        constants = read_constants(xmltree, schema_dict)

    types = {definition.base_type for definition in possible_types}
    if 'float' not in types and \
       'float_expression' not in types and \
       'int' not in types:
        raise ValueError(f"Given attribute name '{attributename}' is not float or int")

    attribs = schema_dict['tag_info'][base_xpath]['attribs']
    _, tag_name = split_off_tag(base_xpath)
    if attributename not in attribs:
        raise ValueError(
            f"The key '{attributename}' is not expected for this version of the input for the '{tag_name}' tag. "
            f'Allowed attributes are: {attribs.original_case.values()}')
    attributename = attribs.original_case[attributename]

    if not xpath.endswith(f'/@{attributename}'):
        xpath = '/@'.join([xpath, attributename])

    stringattribute = eval_xpath(xmltree, xpath, list_return=True)

    tag_xpath, attributename = split_off_attrib(xpath)

    if len(stringattribute) == 0:
        raise ValueError(f"No attribute values found for '{attributename}'. Cannot add number")

    attribvalues, _ = convert_from_xml(stringattribute,
                                       schema_dict,
                                       attributename,
                                       text=False,
                                       constants=constants,
                                       list_return=True)

    if occurrences is not None:
        if not is_sequence(occurrences):
            occurrences = [occurrences]
        try:
            attribvalues = [attribvalues[occ] for occ in occurrences]
        except IndexError as exc:
            raise ValueError('Wrong value for occurrences') from exc

    if mode == 'abs':
        attribvalues = [value + float(add_number) for value in attribvalues]
    elif mode == 'rel':
        attribvalues = [value * float(add_number) for value in attribvalues]

    if 'float' in types or 'float_expression' in types:
        pass
    elif 'int' in types:
        if any(not value.is_integer() for value in attribvalues):
            raise ValueError('You are trying to write a float to an integer attribute')
        attribvalues = [int(value) for value in attribvalues]

    xmltree = xml_set_attrib_value(xmltree,
                                   schema_dict,
                                   tag_xpath,
                                   base_xpath,
                                   attributename,
                                   attribvalues,
                                   occurrences=occurrences)

    return xmltree


def xml_add_number_to_first_attrib(xmltree, schema_dict, xpath, base_xpath, attributename, add_number, mode='abs'):
    """
    Adds a given number to the first occurrence of a attribute value in a xmltree.
    If there are no nodes under the specified xpath an error is raised

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param xpath: a path where to set the attributes
    :param base_xpath: path where to place a new tag without complex syntax ([] conditions and so on)
    :param attributename: the attribute name to change
    :param add_number: number to add/multiply with the old attribute value
    :param mode: str (either `rel` or `abs`).
                 `rel` multiplies the old value with `add_number`
                 `abs` adds the old value and `add_number`

    :raises ValueError: If the attribute is unknown or cannot be float or int
    :raises ValueError: If the evaluation of the old values failed
    :raises ValueError: If a float result is written to a integer attribute

    :returns: xmltree with shifted attribute
    """
    return xml_add_number_to_attrib(xmltree,
                                    schema_dict,
                                    xpath,
                                    base_xpath,
                                    attributename,
                                    add_number,
                                    mode=mode,
                                    occurrences=0)


def xml_set_simple_tag(xmltree, schema_dict, xpath, base_xpath, tag_name, changes, create_parents=False):
    """
    Sets one or multiple `simple` tag(s) in an xmltree. A simple tag can only hold attributes and has no
    subtags.
    If the tag can occur multiple times all existing tags are DELETED and new ones are written.
    If the tag only occurs once it will automatically be created if its missing.

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param xpath: a path where to set the attributes
    :param base_xpath: path where to place a new tag without complex syntax ([] conditions and so on)
    :param tag_name: name of the tag to set
    :param changes: list of dicts or dict with the changes. Elements in list describe multiple tags.
                    Keys in the dictionary correspond to {'attributename': attributevalue}
    :param create_parents: bool optional (default False), if True and the path, where the simple tags are
                           set does not exist it is created

    :returns: xmltree with set simple tags
    """
    from masci_tools.util.xml.xml_setters_basic import xml_delete_tag
    from masci_tools.util.xml.common_functions import check_complex_xpath

    check_complex_xpath(xmltree, base_xpath, xpath)

    tag_info = schema_dict['tag_info'][base_xpath]

    tag_xpath = f'{xpath}/{tag_name}'
    tag_base_xpath = f'{base_xpath}/{tag_name}'

    if tag_name in tag_info['several']:
        #change_dict can either be a list or a dict
        if isinstance(changes, dict):
            changes = [changes]

        # policy: we DELETE all existing tags, and create new ones from the given parameters.
        xml_delete_tag(xmltree, tag_xpath)

        for indx in range(0, len(changes)):
            xml_create_tag_schema_dict(xmltree, schema_dict, xpath, base_xpath, tag_name, create_parents=create_parents)

        for indx, change in enumerate(changes):
            for attrib, value in change.items():
                occurrences = [
                    k * len(changes) + indx
                    for k in range(len(eval_xpath(xmltree, tag_xpath, list_return=True)) // len(changes))
                ]
                xml_set_attrib_value(xmltree,
                                     schema_dict,
                                     tag_xpath,
                                     tag_base_xpath,
                                     attrib,
                                     value,
                                     occurrences=occurrences)
    else:
        if not isinstance(changes, dict):
            raise ValueError(f"Tag '{tag_name}' can only occur once. But 'set_simple_tag' got a list")

        #eval and ggf. create tag
        eval_xpath_create(xmltree, schema_dict, tag_xpath, tag_base_xpath, create_parents=create_parents)

        for attrib, value in changes.items():
            xml_set_attrib_value(xmltree, schema_dict, tag_xpath, tag_base_xpath, attrib, value)

    return xmltree


def xml_set_complex_tag(xmltree, schema_dict, xpath, base_xpath, attributedict, create=False):
    """
    Recursive Function to correctly set tags/attributes for a given tag.
    Goes through the attributedict and decides based on the schema_dict, how the corresponding
    key has to be handled.

    Supports:

        - attributes
        - tags with text only
        - simple tags, i.e. only attributes (can be optional single/multiple)
        - complex tags, will recursively create/modify them

    :param xmltree: an xmltree that represents inp.xml
    :param schema_dict: InputSchemaDict containing all information about the structure of the input
    :param xpath: a path where to set the attributes
    :param base_xpath: path where to place a new tag without complex syntax ([] conditions and so on)
    :param tag_name: name of the tag to set
    :param attributedict: Keys in the dictionary correspond to names of tags and the values are the modifications
                          to do on this tag (attributename, subdict with changes to the subtag, ...)
    :param create: bool optional (default False), if True and the path, where the complex tag is
                   set does not exist it is created

    :returns: xmltree with changes to the complex tag
    """
    from masci_tools.util.xml.xml_setters_basic import xml_delete_tag
    from masci_tools.util.xml.common_functions import check_complex_xpath, split_off_tag

    check_complex_xpath(xmltree, base_xpath, xpath)

    tag_info = schema_dict['tag_info'][base_xpath]
    _, tag_name = split_off_tag(base_xpath)

    if create:
        #eval complex tag and ggf create
        eval_xpath_create(xmltree, schema_dict, xpath, base_xpath, create_parents=True)

    for key, val in attributedict.items():

        if key not in tag_info['complex'] | tag_info['simple'] | tag_info['attribs']:
            raise ValueError(
                f"The key '{key}' is not expected for this version of the input for the '{tag_name}' tag. "
                f"Allowed tags are: {sorted((tag_info['complex']|tag_info['simple']).original_case.values())}"
                f"Allowed attributes are: {sorted(tag_info['attribs'].original_case.values())}")

        key = (tag_info['complex'] | tag_info['simple'] | tag_info['attribs']).original_case[key]

        sub_xpath = f'{xpath}/{key}'
        sub_base_xpath = f'{base_xpath}/{key}'
        if key in tag_info['attribs']:
            xml_set_attrib_value(xmltree, schema_dict, xpath, base_xpath, key, val, create=create)

        elif key in tag_info['text']:
            xml_set_text(xmltree, schema_dict, sub_xpath, sub_base_xpath, val, create=create)

        elif key in tag_info['simple']:
            xml_set_simple_tag(xmltree, schema_dict, xpath, base_xpath, key, val, create_parents=create)

        elif key not in tag_info['several']:  #Complex tag but only one (electronConfig)

            # eval and ggf create tag at right place.
            eval_xpath_create(xmltree, schema_dict, sub_xpath, sub_base_xpath, create_parents=create)

            xmltree = xml_set_complex_tag(xmltree, schema_dict, sub_xpath, sub_base_xpath, val, create=create)

        else:
            # policy: we DELETE all existing tags, and create new ones from the given parameters.
            xml_delete_tag(xmltree, sub_xpath)

            if isinstance(val, dict):
                val = [val]

            for indx in range(0, len(val)):
                xml_create_tag_schema_dict(xmltree, schema_dict, xpath, base_xpath, key, create_parents=create)

            for indx, tagdict in enumerate(val):
                for k in range(len(eval_xpath(xmltree, sub_xpath, list_return=True)) // len(val)):
                    current_elem_xpath = f'{sub_xpath}[{k*len(val)+indx+1}]'
                    xmltree = xml_set_complex_tag(xmltree,
                                                  schema_dict,
                                                  current_elem_xpath,
                                                  sub_base_xpath,
                                                  tagdict,
                                                  create=create)

    return xmltree



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
This module contains helper functions for extracting information easily from the
schema_dicts defined for the Fleur input/output

Also provides convienient functions to use just a attribute name for extracting the
attribute from the right place in the given etree
"""
from masci_tools.util.parse_tasks_decorators import register_parsing_function
from masci_tools.util.lockable_containers import LockableList
from lxml import etree


def _find_paths(schema_dict, name, entries, contains=None, not_contains=None):
    """
    Find all paths in the schema_dict in the given entries for the given name
    and matching the contains/not_contains criteria

    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param contains: str or list of str, this string has to be in the final path
    :param not_contains: str or list of str, this string has to NOT be in the final path

    :returns: list of str, found xpaths matching the criteria
    """

    if contains is None:
        contains = []
    elif not isinstance(contains, (list, set)):
        contains = [contains]

    if not_contains is None:
        not_contains = []
    elif not isinstance(not_contains, (list, set)):
        not_contains = [not_contains]

    path_list = []
    for entry in entries:
        if name in schema_dict[entry]:
            entry_paths = schema_dict[entry][name]

            if not isinstance(entry_paths, LockableList):
                entry_paths = [entry_paths]
            else:
                entry_paths = entry_paths.get_unlocked()

            invalid_paths = set()
            for phrase in contains:
                for xpath in entry_paths:
                    if phrase not in xpath:
                        invalid_paths.add(xpath)

            for phrase in not_contains:
                for xpath in entry_paths:
                    if phrase in xpath:
                        invalid_paths.add(xpath)

            for invalid in invalid_paths:
                entry_paths.remove(invalid)

            path_list += entry_paths

    return path_list


def get_tag_xpath(schema_dict, name, contains=None, not_contains=None):
    """
    Tries to find a unique path from the schema_dict based on the given name of the tag
    and additional further specifications

    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param contains: str or list of str, this string has to be in the final path
    :param not_contains: str or list of str, this string has to NOT be in the final path

    :returns: str, xpath for the given tag

    :raises ValueError: If no unique path could be found
    """

    possible_lists = ['tag_paths']

    if 'iteration_tag_paths' in schema_dict:
        possible_lists += ['iteration_tag_paths']

    paths = _find_paths(schema_dict, name, possible_lists, contains=contains, not_contains=not_contains)

    if len(paths) == 1:
        return paths[0]
    elif len(paths) == 0:
        raise ValueError(f'The tag {name} has no possible paths with the current specification.\n'
                         f'contains: {contains}, not_contains: {not_contains}')
    else:
        raise ValueError(f'The tag {name} has multiple possible paths with the current specification.\n'
                         f'contains: {contains}, not_contains: {not_contains} \n'
                         f'These are possible: {paths}')


def get_relative_tag_xpath(schema_dict, name, root_tag, contains=None, not_contains=None):
    """
    Tries to find a unique relative path from the schema_dict based on the given name of the tag
    name of the root, from which the path should be relative and additional further specifications

    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param root_tag: str, name of the tag from which the path should be relative
    :param contains: str or list of str, this string has to be in the final path
    :param not_contains: str or list of str, this string has to NOT be in the final path

    :returns: str, xpath for the given tag

    :raises ValueError: If no unique path could be found
    """
    from masci_tools.util.xml.common_functions import abs_to_rel_xpath

    possible_lists = ['tag_paths']

    if 'iteration_tag_paths' in schema_dict:
        possible_lists += ['iteration_tag_paths']

    #The paths have to include the root_tag
    if contains is None:
        contains = [root_tag]
    else:
        contains = set(contains)
        contains.add(root_tag)

    paths = _find_paths(schema_dict, name, possible_lists, contains=contains, not_contains=not_contains)

    rel_paths = set()
    for xpath in paths:
        rel_paths.add(abs_to_rel_xpath(xpath, root_tag))

    if len(rel_paths) == 1:
        return rel_paths.pop()
    elif len(rel_paths) == 0:
        raise ValueError(f'The tag {name} has no possible relative paths with the current specification.\n'
                         f'contains: {contains}, not_contains: {not_contains}, root_tag {root_tag}')
    else:
        raise ValueError(f'The tag {name} has multiple possible relative paths with the current specification.\n'
                         f'contains: {contains}, not_contains: {not_contains}, root_tag {root_tag} \n'
                         f'These are possible: {rel_paths}')


def get_attrib_xpath(schema_dict, name, contains=None, not_contains=None, exclude=None, tag_name=None):
    """
    Tries to find a unique path from the schema_dict based on the given name of the attribute
    and additional further specifications

    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the attribute
    :param root_tag: str, name of the tag from which the path should be relative
    :param contains: str or list of str, this string has to be in the final path
    :param not_contains: str or list of str, this string has to NOT be in the final path
    :param exclude: list of str, here specific types of attributes can be excluded
                    valid values are: settable, settable_contains, other
    :param tag_name: str, if given this name will be used to find a path to a tag with the
                     same name in :py:func:`get_tag_xpath()`

    :returns: str, xpath to the tag with the given attribute

    :raises ValueError: If no unique path could be found
    """

    if tag_name is not None:
        tag_xpath = get_tag_xpath(schema_dict, tag_name, contains=contains, not_contains=not_contains)

        err_msg = f'No attribute {name} found at tag {tag_name}'
        if tag_xpath in schema_dict['tag_info']:
            if name not in schema_dict['tag_info'][tag_xpath]['attribs']:
                raise ValueError(err_msg)
        else:
            if name not in schema_dict['iteration_tag_info'][tag_xpath]['attribs']:
                raise ValueError(err_msg)
        return f'{tag_xpath}/@{name}'

    possible_lists = ['unique_attribs', 'unique_path_attribs', 'other_attribs']
    output = False
    if 'iteration_unique_attribs' in schema_dict:
        #outputschema
        output = True
        possible_lists += ['iteration_unique_attribs', 'iteration_unique_path_attribs', 'iteration_other_attribs']

    if exclude is not None:
        for list_name in exclude:
            possible_lists.remove(f'{list_name}_attribs')
            if output:
                possible_lists.remove(f'iteration_{list_name}_attribs')

    paths = _find_paths(schema_dict, name, possible_lists, contains=contains, not_contains=not_contains)

    if len(paths) == 1:
        return paths[0]
    elif len(paths) == 0:
        raise ValueError(f'The attrib {name} has no possible paths with the current specification.\n'
                         f'contains: {contains}, not_contains: {not_contains}, exclude {exclude}')
    else:
        raise ValueError(f'The attrib {name} has multiple possible paths with the current specification.\n'
                         f'contains: {contains}, not_contains: {not_contains}, exclude {exclude}\n'
                         f'These are possible: {paths}')


def get_relative_attrib_xpath(schema_dict,
                              name,
                              root_tag,
                              contains=None,
                              not_contains=None,
                              exclude=None,
                              tag_name=None):
    """
    Tries to find a unique relative path from the schema_dict based on the given name of the attribute
    name of the root, from which the path should be relative and additional further specifications

    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the attribute
    :param contains: str or list of str, this string has to be in the final path
    :param not_contains: str or list of str, this string has to NOT be in the final path
    :param exclude: list of str, here specific types of attributes can be excluded
                    valid values are: settable, settable_contains, other
    :param tag_name: str, if given this name will be used to find a path to a tag with the
                     same name in :py:func:`get_relative_tag_xpath()`

    :returns: str, xpath for the given tag

    :raises ValueError: If no unique path could be found
    """
    from masci_tools.util.xml.common_functions import abs_to_rel_xpath

    if tag_name is not None:
        tag_xpath = get_relative_tag_xpath(schema_dict,
                                           tag_name,
                                           root_tag,
                                           contains=contains,
                                           not_contains=not_contains)

        tag_info = get_tag_info(schema_dict,
                                tag_name,
                                path_return=False,
                                multiple_paths=True,
                                contains=contains,
                                not_contains=not_contains)

        err_msg = f'No attribute {name} found at tag {tag_name}'
        if name not in tag_info['attribs']:
            raise ValueError(err_msg)

        if tag_xpath.endswith('/'):
            return f'{tag_xpath}@{name}'
        else:
            return f'{tag_xpath}/@{name}'

    possible_lists = ['unique_attribs', 'unique_path_attribs', 'other_attribs']
    output = False
    if 'iteration_unique_attribs' in schema_dict:
        #outputschema
        output = True
        possible_lists += ['iteration_unique_attribs', 'iteration_unique_path_attribs', 'iteration_other_attribs']

    if exclude is not None:
        for list_name in exclude:
            possible_lists.remove(f'{list_name}_attribs')
            if output:
                possible_lists.remove(f'iteration_{list_name}_attribs')

    #The paths have to include the root_tag
    if contains is None:
        contains = [root_tag]
    else:
        contains = set(contains)
        contains.add(root_tag)

    paths = _find_paths(schema_dict, name, possible_lists, contains=contains, not_contains=not_contains)

    rel_paths = set()
    for xpath in paths:
        rel_paths.add(abs_to_rel_xpath(xpath, root_tag))

    if len(rel_paths) == 1:
        return rel_paths.pop()
    elif len(rel_paths) == 0:
        raise ValueError(f'The attrib {name} has no possible relative paths with the current specification.\n'
                         f'contains: {contains}, not_contains: {not_contains}, root_tag {root_tag}')
    else:
        raise ValueError(f'The attrib {name} has multiple possible relative paths with the current specification.\n'
                         f'contains: {contains}, not_contains: {not_contains}, root_tag {root_tag} \n'
                         f'These are possible: {rel_paths}')


def get_tag_info(schema_dict,
                 name,
                 contains=None,
                 not_contains=None,
                 path_return=True,
                 convert_to_builtin=False,
                 multiple_paths=False,
                 parent=False):
    """
    Tries to find a unique path from the schema_dict based on the given name of the tag
    and additional further specifications and returns the tag_info entry for this tag

    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param contains: str or list of str, this string has to be in the final path
    :param not_contains: str or list of str, this string has to NOT be in the final path
    :param path_return: bool, if True the found path will be returned alongside the tag_info
    :param convert_to_builtin: bool, if True the CaseInsensitiveFrozenSets are converetd to normal sets
                               with the rigth case of the attributes
    :param multiple_paths: bool, if True mulitple paths are allowed to match as long as they have the same tag_info
    :param parent: bool, if True the tag_info for the parent of the tag is returned

    :returns: dict, tag_info for the found xpath
    :returns: str, xpath to the tag if `path_return=True`
    """
    import copy
    from masci_tools.util.case_insensitive_dict import CaseInsensitiveFrozenSet, CaseInsensitiveDict
    from masci_tools.util.xml.common_functions import split_off_tag

    if multiple_paths:
        possible_lists = ['tag_paths']

        if 'iteration_tag_paths' in schema_dict:
            possible_lists += ['iteration_tag_paths']

        paths = _find_paths(schema_dict, name, possible_lists, contains=contains, not_contains=not_contains)
    else:
        paths = [get_tag_xpath(schema_dict, name, contains=contains, not_contains=not_contains)]

    EMPTY_TAG_INFO = CaseInsensitiveDict({
        'attribs': CaseInsensitiveFrozenSet(),
        'optional_attribs': {},
        'optional': CaseInsensitiveFrozenSet(),
        'order': [],
        'several': CaseInsensitiveFrozenSet(),
        'simple': CaseInsensitiveFrozenSet(),
        'complex': CaseInsensitiveFrozenSet(),
        'text': CaseInsensitiveFrozenSet()
    })
    EMPTY_TAG_INFO.freeze()

    tag_info = None
    for path in paths:

        if parent:
            path, _ = split_off_tag(path)

        if path in schema_dict['tag_info']:
            entry = schema_dict['tag_info'][path]
        elif 'iteration_tag_info' in schema_dict:
            if path in schema_dict['iteration_tag_info']:
                entry = schema_dict['iteration_tag_info'][path]
            else:
                entry = EMPTY_TAG_INFO
        else:
            entry = EMPTY_TAG_INFO

        if tag_info is not None:
            if entry != tag_info:
                raise ValueError(f'Differing tag_info for the found paths {paths}')
        else:
            tag_info = entry

    if not multiple_paths:
        paths = paths[0]

    if convert_to_builtin:
        tag_info = {
            key: set(val.original_case.values()) if isinstance(val, CaseInsensitiveFrozenSet) else val
            for key, val in tag_info.items()
        }

    if path_return:
        return tag_info, paths
    else:
        return tag_info


def read_constants(root, schema_dict, logger=None):
    """
    Reads in the constants defined in the inp.xml
    and returns them combined with the predefined constants from
    fleur as a dictionary

    :param root: root of the etree of the inp.xml file
    :param schema_dict: schema_dictionary of the version of the file to read (inp.xml or out.xml)
    :param logger: logger object for logging warnings, errors

    :return: a python dictionary with all defined constants
    """
    from masci_tools.util.constants import FLEUR_DEFINED_CONSTANTS
    import copy
    import warnings

    defined_constants = copy.deepcopy(FLEUR_DEFINED_CONSTANTS)

    try:
        tag_exists(root, schema_dict, 'constant')
    except ValueError as err:
        if 'no possible' in str(err):
            warnings.warn('Cannot extract custom constants for the given root. Assuming defaults')
            return defined_constants
        else:
            raise

    if not tag_exists(root, schema_dict, 'constant', logger=logger):  #Avoid warnings for empty constants
        return defined_constants

    constants = evaluate_tag(root, schema_dict, 'constant', defined_constants, logger=logger)

    if constants['name'] is not None:
        if not isinstance(constants['name'], list):
            constants = {key: [val] for key, val in constants.items()}
        for name, value in zip(constants['name'], constants['value']):
            if name not in defined_constants:
                defined_constants[name] = value
            else:
                if logger is not None:
                    logger.error('Ambiguous definition of constant %s', name)
                raise KeyError(f'Ambiguous definition of constant {name}')

    return defined_constants


@register_parsing_function('attrib')
def evaluate_attribute(node, schema_dict, name, constants=None, logger=None, **kwargs):
    """
    Evaluates the value of the attribute based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the attribute
    :param constants: dict, contains the defined constants
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised

    Kwargs:
        :param tag_name: str, name of the tag where the attribute should be parsed
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param exclude: list of str, here specific types of attributes can be excluded
                        valid values are: settable, settable_contains, other
        :param list_return: if True, the returned quantity is always a list even if only one element is in it
        :param optional: bool, if True and no logger given none or an empty list is returned

    :returns: list or single value, converted in convert_xml_attribute
    """
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.converters import convert_from_xml

    list_return = kwargs.pop('list_return', False)
    optional = kwargs.pop('optional', False)

    attrib_xpath = None
    if isinstance(node, etree._Element):
        if node.tag != schema_dict['root_tag'] and node.tag != 'iteration':
            attrib_xpath = get_relative_attrib_xpath(schema_dict, name, node.tag, **kwargs)

    if attrib_xpath is None:
        attrib_xpath = get_attrib_xpath(schema_dict, name, **kwargs)

    stringattribute = eval_xpath(node, attrib_xpath, logger=logger, list_return=True)

    if len(stringattribute) == 0:
        if logger is None:
            if not optional:
                raise ValueError(f'No values found for attribute {name}')
        else:
            logger.warning('No values found for attribute %s', name)
        if list_return:
            return []
        else:
            return None

    converted_value, suc = convert_from_xml(stringattribute,
                                            schema_dict,
                                            name,
                                            text=False,
                                            constants=constants,
                                            logger=logger,
                                            list_return=list_return)

    if not suc:
        if logger is None:
            raise ValueError(f'Failed to evaluate attribute {name}, Got value: {stringattribute}')
        else:
            logger.warning('Failed to evaluate attribute %s, Got value: %s', name, stringattribute)

    return converted_value


@register_parsing_function('text')
def evaluate_text(node, schema_dict, name, constants=None, logger=None, **kwargs):
    """
    Evaluates the text of the tag based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param constants: dict, contains the defined constants
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param list_return: if True, the returned quantity is always a list even if only one element is in it
        :param optional: bool, if True and no logger given none or an empty list is returned

    :returns: list or single value, converted in convert_xml_text
    """
    from masci_tools.util.xml.common_functions import eval_xpath
    from masci_tools.util.xml.converters import convert_from_xml

    list_return = kwargs.pop('list_return', False)
    optional = kwargs.pop('optional', False)

    tag_xpath = None
    if isinstance(node, etree._Element):
        if node.tag != schema_dict['root_tag'] and node.tag != 'iteration':
            tag_xpath = get_relative_tag_xpath(schema_dict, name, node.tag, **kwargs)

    if tag_xpath is None:
        tag_xpath = get_tag_xpath(schema_dict, name, **kwargs)

    stringtext = eval_xpath(node, f'{tag_xpath}/text()', logger=logger, list_return=True)

    for text in stringtext.copy():
        if text.strip() == '':
            stringtext.remove(text)

    if len(stringtext) == 0:
        if logger is None:
            if not optional:
                raise ValueError(f'No text found for tag {name}')
        else:
            logger.warning('No text found for tag %s', name)
        if list_return:
            return []
        else:
            return None

    converted_value, suc = convert_from_xml(stringtext,
                                            schema_dict,
                                            name,
                                            text=True,
                                            constants=constants,
                                            logger=logger,
                                            list_return=list_return)

    if not suc:
        if logger is None:
            raise ValueError(f'Failed to evaluate text for tag {name}, Got text: {stringtext}')
        else:
            logger.warning('Failed to evaluate text for tag %s, Got text: %s', name, stringtext)

    return converted_value


@register_parsing_function('allAttribs', all_attribs_keys=True)
def evaluate_tag(node, schema_dict, name, constants=None, logger=None, subtags=False, text=True, **kwargs):
    """
    Evaluates all attributes of the tag based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param constants: dict, contains the defined constants
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised
    :param subtags: optional bool, if True the subtags of the given tag are evaluated
    :param text: optional bool, if True the text of the tag is also parsed

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param only_required: bool (optional, default False), if True only required attributes are parsed
        :param ignore: list of str (optional), attributes not to parse
        :param list_return: if True, the returned quantity is always a list even if only one element is in it
        :param strict_missing_error: if True, and no logger is given an error is raised if any attribute is not found

    :returns: dict, with attribute values converted via convert_xml_attribute
    """
    from masci_tools.util.xml.common_functions import eval_xpath, split_off_tag
    from masci_tools.util.xml.converters import convert_from_xml

    only_required = kwargs.pop('only_required', False)
    strict_missing_error = kwargs.pop('strict_missing_error', False)
    ignore = kwargs.pop('ignore', None)
    list_return = kwargs.pop('list_return', False)

    tag_xpath = None
    if isinstance(node, etree._Element):
        if node.tag != schema_dict['root_tag'] and node.tag != 'iteration':
            kwargs['contains'] = set(kwargs.get('contains', []))
            kwargs['contains'].add(node.tag)
            tag_xpath = get_relative_tag_xpath(schema_dict, name, node.tag, **kwargs)

    if tag_xpath is None:
        tag_xpath = get_tag_xpath(schema_dict, name, **kwargs)

    #Which attributes are expected
    tags = set()
    optional_tags = set()
    try:
        tag_info = get_tag_info(schema_dict, name, path_return=False, multiple_paths=True, **kwargs)
        attribs = tag_info['attribs']
        optional = tag_info['optional_attribs']
        if subtags:
            tags = tag_info['simple'] | tag_info['complex']
            optional_tags = tag_info['optional']
    except ValueError as err:
        if logger is None:
            raise ValueError(f'Failed to evaluate attributes from tag {name}: '
                             'No attributes to parse either the tag does not '
                             'exist or it has no attributes') from err
        else:
            logger.exception(
                'Failed to evaluate attributes from tag %s: '
                'No attributes to parse either the tag does not '
                'exist or it has no attributes', name)
        attribs = set()
        optional = set()
        tags = set()
        optional_tags = set()

    if only_required:
        attribs = attribs.difference(optional)
        tags = tags.difference(optional_tags)

    if ignore:
        attribs = attribs.difference(ignore)
        tags = tags.difference(ignore)

    parse_text = name in schema_dict['text_tags'] and text

    if not attribs and not parse_text and not tags:
        if subtags:
            return {}
        if logger is None:
            raise ValueError(f'Failed to evaluate attributes from tag {name}: '
                             'No attributes to parse either the tag does not '
                             'exist or it has no attributes')
        else:
            logger.error(
                'Failed to evaluate attributes from tag %s: '
                'No attributes to parse either the tag does not '
                'exist or it has no attributes', name)
    else:
        attribs = sorted(list(attribs.original_case.values()))

    out_dict = {}

    for attrib in attribs:

        stringattribute = eval_xpath(node, f'{tag_xpath}/@{attrib}', logger=logger, list_return=True)

        if len(stringattribute) == 0:
            if logger is None:
                if strict_missing_error and attrib not in optional:
                    raise ValueError(f'No values found for attribute {attrib} at tag {name}')
            else:
                logger.warning('No values found for attribute %s at tag %s', attrib, name)
            if list_return:
                out_dict[attrib] = []
            else:
                out_dict[attrib] = None
            continue

        out_dict[attrib], suc = convert_from_xml(stringattribute,
                                                 schema_dict,
                                                 attrib,
                                                 text=False,
                                                 constants=constants,
                                                 logger=logger,
                                                 list_return=list_return)
        if not suc:
            if logger is None:
                raise ValueError(f'Failed to evaluate attribute {attrib}, Got value: {stringattribute}')
            else:
                logger.warning('Failed to evaluate attribute %s, Got value: %s', attrib, stringattribute)

    if parse_text:

        _, name = split_off_tag(tag_xpath)
        stringtext = eval_xpath(node, f'{tag_xpath}/text()', logger=logger, list_return=True)

        for textval in stringtext.copy():
            if textval.strip() == '':
                stringtext.remove(text)

        if len(stringtext) == 0:
            if logger is None:
                if not optional:
                    raise ValueError(f'No text found for tag {name}')
            else:
                logger.warning('No text found for tag %s', name)
            if list_return:
                out_dict[name] = []
            else:
                out_dict[name] = None

        out_dict[name], suc = convert_from_xml(stringtext,
                                               schema_dict,
                                               name,
                                               text=True,
                                               constants=constants,
                                               logger=logger,
                                               list_return=list_return)

    if subtags:
        for tag in tags:
            if tag in out_dict:
                if logger is None:
                    raise ValueError(f'Conflicting key {tag}: ' 'Key is already in the output dictionary')
                else:
                    logger.error('Conflicting key %s: ' 'Key is already in the output dictionary', tag)
            out_dict[tag] = []

        sub_nodes = eval_xpath(node, tag_xpath, logger=logger, list_return=True)
        for sub_node in sub_nodes:
            for tag in tags:
                if tag_exists(sub_node, schema_dict, tag):
                    out_dict[tag].append(
                        evaluate_tag(sub_node,
                                     schema_dict,
                                     tag,
                                     constants=constants,
                                     logger=logger,
                                     subtags=True,
                                     ignore=ignore,
                                     only_required=only_required,
                                     strict_missing_error=strict_missing_error,
                                     list_return=list_return,
                                     text=text))

        for tag in tags:
            for sub_dict in out_dict[tag]:
                if not sub_dict:
                    out_dict[tag].remove(sub_dict)
                elif len(sub_dict) == 1 and tag in sub_dict:
                    out_dict[tag] = sub_dict[tag]

        for tag in tags:
            if len(out_dict[tag]) == 1:
                out_dict[tag] = out_dict[tag][0]
            elif len(out_dict[tag]) == 0:
                out_dict.pop(tag)

    return out_dict


@register_parsing_function('singleValue', all_attribs_keys=True)
def evaluate_single_value_tag(node, schema_dict, name, constants=None, logger=None, **kwargs):
    """
    Evaluates the value and unit attribute of the tag based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param constants: dict, contains the defined constants
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param only_required: bool (optional, default False), if True only required attributes are parsed
        :param ignore: list of str (optional), attributes not to parse
        :param list_return: if True, the returned quantity is always a list even if only one element is in it
        :param strict_missing_error: if True, and no logger is given an error is raised if any attribute is not found

    :returns: value and unit, both converted in convert_xml_attribute
    """

    only_required = kwargs.get('only_required', False)
    ignore = kwargs.get('ignore', [])

    value_dict = evaluate_tag(node, schema_dict, name, constants=constants, logger=logger, **kwargs)

    if value_dict.get('value') is None:
        if logger is None:
            raise ValueError(f'Failed to evaluate singleValue from tag {name}: ' "Has no 'value' attribute")
        else:
            logger.warning('Failed to evaluate singleValue from tag %s: ' "Has no 'value' attribute", name)

    if value_dict.get('units') is None and not only_required and 'units' not in ignore:
        if logger is None:
            raise ValueError(f'Failed to evaluate singleValue from tag {name}: ' "Has no 'units' attribute")
        else:
            logger.warning('Failed to evaluate singleValue from tag %s: ' "Has no 'units' attribute", name)

    return value_dict


@register_parsing_function('parentAttribs', all_attribs_keys=True)
def evaluate_parent_tag(node, schema_dict, name, constants=None, logger=None, **kwargs):
    """
    Evaluates all attributes of the parent tag based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param constants: dict, contains the defined constants
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param only_required: bool (optional, default False), if True only required attributes are parsed
        :param ignore: list of str (optional), attributes not to parse
        :param list_return: if True, the returned quantity is always a list even if only one element is in it
        :param strict_missing_error: if True, and no logger is given an error is raised if any attribute is not found

    :returns: dict, with attribute values converted via convert_xml_attribute
    """
    from masci_tools.util.xml.common_functions import eval_xpath, get_xml_attribute
    from masci_tools.util.xml.converters import convert_from_xml

    strict_missing_error = kwargs.pop('strict_missing_error', False)
    list_return = kwargs.pop('list_return', False)
    only_required = kwargs.pop('only_required', False)
    ignore = kwargs.pop('ignore', None)

    tag_xpath = None
    if isinstance(node, etree._Element):
        if node.tag != schema_dict['root_tag'] and node.tag != 'iteration':
            kwargs['contains'] = set(kwargs.get('contains', []))
            kwargs['contains'].add(node.tag)
            tag_xpath = get_relative_tag_xpath(schema_dict, name, node.tag, **kwargs)

    if tag_xpath is None:
        tag_xpath = get_tag_xpath(schema_dict, name, **kwargs)

    #Which attributes are expected
    try:
        tag_info = get_tag_info(schema_dict, name, path_return=False, multiple_paths=True, parent=True, **kwargs)
        attribs = tag_info['attribs']
        optional = tag_info['optional_attribs']
    except ValueError as err:
        if logger is None:
            raise ValueError(f'Failed to evaluate attributes from parent tag of {name}: '
                             'No attributes to parse either the tag does not '
                             'exist or it has no attributes') from err
        else:
            logger.exception(
                'Failed to evaluate attributes from parent tag of %s: '
                'No attributes to parse either the tag does not '
                'exist or it has no attributes', name)
        attribs = set()
        optional = set()

    if only_required:
        attribs = attribs.difference(optional)

    if ignore is not None:
        attribs = attribs.difference(ignore)

    if not attribs:
        if logger is None:
            raise ValueError(f'Failed to evaluate attributes from parent tag of {name}: '
                             'No attributes to parse either the tag does not '
                             'exist or it has no attributes')
        else:
            logger.error(
                'Failed to evaluate attributes from parent tag of %s: '
                'No attributes to parse either the tag does not '
                'exist or it has no attributes', name)
    else:
        attribs = sorted(list(attribs.original_case.values()))

    elems = eval_xpath(node, tag_xpath, logger=logger, list_return=True)

    out_dict = {}
    for attrib in attribs:
        out_dict[attrib] = []

    for elem in elems:
        parent = elem.getparent()
        for attrib in attribs:

            stringattribute = get_xml_attribute(parent, attrib, logger=logger)

            if stringattribute is None:
                if logger is None:
                    if strict_missing_error and attrib not in optional:
                        raise ValueError(f'No values found for attribute {attrib} for parent tag of {name}')
                else:
                    logger.warning('No values found for attribute %s for parent tag of %s', attrib, name)
                out_dict[attrib].append(None)
                continue

            value, suc = convert_from_xml(stringattribute,
                                          schema_dict,
                                          attrib,
                                          text=attrib in tag_info['text'],
                                          constants=constants,
                                          logger=logger,
                                          list_return=list_return)

            out_dict[attrib].append(value)

            if not suc:
                if logger is None:
                    raise ValueError(f'Failed to evaluate attribute {attrib}, Got value: {stringattribute}')
                else:
                    logger.warning('Failed to evaluate attribute %s, Got value: %s', attrib, stringattribute)

    if all(len(x) == 1 for x in out_dict.values()) and not list_return:
        out_dict = {key: val[0] for key, val in out_dict.items()}

    return out_dict


@register_parsing_function('attrib_exists')
def attrib_exists(node, schema_dict, name, logger=None, **kwargs):
    """
    Evaluates whether the attribute exists in the xmltree based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised

    Kwargs:
        :param tag_name: str, name of the tag where the attribute should be parsed
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param exclude: list of str, here specific types of attributes can be excluded
                        valid values are: settable, settable_contains, other

    :returns: bool, True if any tag with the attribute exists
    """
    from masci_tools.util.xml.common_functions import eval_xpath, split_off_attrib

    attrib_xpath = None
    if isinstance(node, etree._Element):
        if node.tag != schema_dict['root_tag'] and node.tag != 'iteration':
            attrib_xpath = get_relative_attrib_xpath(schema_dict, name, node.tag, **kwargs)

    if attrib_xpath is None:
        attrib_xpath = get_attrib_xpath(schema_dict, name, **kwargs)

    tag_xpath, attrib_name = split_off_attrib(attrib_xpath)

    tags = eval_xpath(node, tag_xpath, logger=logger, list_return=True)

    return any(attrib_name in tag.attrib for tag in tags)


@register_parsing_function('exists')
def tag_exists(node, schema_dict, name, logger=None, **kwargs):
    """
    Evaluates whether the tag exists in the xmltree based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path

    :returns: bool, True if any nodes with the path exist
    """
    return get_number_of_nodes(node, schema_dict, name, logger=logger, **kwargs) != 0


@register_parsing_function('numberNodes')
def get_number_of_nodes(node, schema_dict, name, logger=None, **kwargs):
    """
    Evaluates the number of occurences of the tag in the xmltree based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path

    :returns: bool, True if any nodes with the path exist
    """
    return len(eval_simple_xpath(node, schema_dict, name, logger=logger, list_return=True, **kwargs))


def eval_simple_xpath(node, schema_dict, name, logger=None, **kwargs):
    """
    Evaluates a simple xpath expression of the tag in the xmltree based on the given name
    and additional further specifications with the available type information

    :param node: etree Element, on which to execute the xpath evaluations
    :param schema_dict: dict, containing all the path information and more
    :param name: str, name of the tag
    :param logger: logger object for logging warnings, errors, if not provided all errors will be raised

    Kwargs:
        :param contains: str, this string has to be in the final path
        :param not_contains: str, this string has to NOT be in the final path
        :param list_return: bool, if True a list is always returned

    :returns: etree Elements obtained via the simple xpath expression
    """
    from masci_tools.util.xml.common_functions import eval_xpath

    list_return = kwargs.pop('list_return', False)

    tag_xpath = None
    if isinstance(node, etree._Element):
        if node.tag != schema_dict['root_tag'] and node.tag != 'iteration':
            tag_xpath = get_relative_tag_xpath(schema_dict, name, node.tag, **kwargs)

    if tag_xpath is None:
        tag_xpath = get_tag_xpath(schema_dict, name, **kwargs)

    return eval_xpath(node, tag_xpath, logger=logger, list_return=list_return)



# -*- coding: utf-8 -*-
"""
This module collects functions for calculating properties with the greens functions
calculated by Fleur. At the moment the following are implemented:

   * Calculating Heisenberg J_0 (spin stiffness) from onsite Green's functions
   * Calculating Heisenberg J_ij exchange constants from intersite Green's functions
   * Calculating the hybridization function from onsite Greens functions
"""
from .greensfunction import intersite_shells, intersite_shells_from_file

import numpy as np
from scipy import constants
from collections import defaultdict
from typing import Dict, List


def calculate_heisenberg_jij(hdffileORgreensfunctions, reference_atom, onsite_delta, show=False):
    r"""
    Calculate the Heisenberg exchange constants form Green's functions using the formula

    .. math::
        J_{ij} = \frac{1}{4\pi} \mathrm{Im}\ \mathrm{Tr_L} \int_{-\infty}^{E_F}\!\mathrm{dz} \Delta_iG_{ij}(z)\Delta_jG_{ji}(z)

    :param hdffileORgreensfunctions: either pat/file-like object for the ``greensf.hdf`` file to use or list of :py:class:`~masci_tools.tools.greensfunction.GreensFunction`
    :param reference_atom: integer index of the atom to calculate the Jijs from
    :param onsite_delta: List of floats containing the onsite exchange splitting for each atom type and l-channel
    :param show: bool if True additional information about the used Greens functions is printed out

    :returns: dict mapping the distances to all the calculated Jijs for that distance
    """

    if isinstance(hdffileORgreensfunctions, list):
        shells = intersite_shells(hdffileORgreensfunctions, reference_atom, show=show)
    else:
        shells = intersite_shells_from_file(hdffileORgreensfunctions, reference_atom, show=show)

    jij_constants: Dict[float, List[float]] = defaultdict(list)

    for dist, g1, g2 in shells:

        dist = round(dist, 12)

        jij = 0.0
        gij = g1.energy_dependence(spin=1, both_contours=True)
        gji = g2.energy_dependence(spin=2, both_contours=True)

        for weight, gijz, gjiz in zip(g1.weights, gij, gji):

            jij += 1/(8.0*np.pi*1j) * onsite_delta[g1.atomType-1,g1.l] * onsite_delta[g1.atomTypep-1,g1.l]\
                  * (weight * np.trace(gijz[...,0] @ gjiz[...,0]) \
                     - weight.conj() * np.trace(gijz[...,1] @ gjiz[...,1]))

        jij_constants[dist].append(jij.real * 1000.0)
        if show:
            print(f'distance: {dist}; J_{g1.l}{g1.lp}: {jij.real*1000.0} meV')

    if show:
        print(f'J_0 calculated as sum of J_ij constants: {sum(np.sum(jijs) for jijs in jij_constants.values())} meV')

    return dict(jij_constants)


def calculate_heisenberg_j0(g, onsite_delta, show=False):

    g_up = g.energy_dependence(spin=1, both_contours=True)
    g_dn = g.energy_dependence(spin=2, both_contours=True)

    j0 = 0.0
    for weight, g_upz, g_dnz in zip(g.weights, g_up, g_dn):

        j0 += 1/(8.0*np.pi*1j) * (\
              weight *        (onsite_delta * (np.trace(g_upz[...,0])-np.trace(g_dnz[...,0])) \
                                + onsite_delta**2 * np.trace(g_upz[...,0].dot(g_dnz[...,0]))) \
             -weight.conj() * (onsite_delta * (np.trace(g_upz[...,1])-np.trace(g_dnz[...,1])) \
                                + onsite_delta**2 * np.trace(g_upz[...,1].dot(g_dnz[...,1]))))

    j0 = j0.real * 1000.0
    if show:
        print(f'Effective exchange Interaction: {j0:.4f} meV')
        print('Curie Temperature: {:.4f} K'.format(2.0/3.0*j0 * \
              1/constants.value('Boltzmann constant in eV/K')*1/1000.0)) #1/1000 to convert to meV

    return j0


def calculate_hybridization(g):

    gz = g.trace_energy_dependence()
    delta = 1 / (2 * g.l + 1) * np.linalg.inv(gz)
    return delta.real



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools    #
# For further information on the license, see the LICENSE.txt file            #
# For further information please visit http://www.flapw.de or                 #
#                                                                             #
###############################################################################
"""
This modules provides common plotting functions dispatching to
different plotting backends. At the moment the following backends are used:

* ``matplotlib`` ('mpl', 'matplotlib')
* ``bokeh`` ('bokeh')

The underlying plotting routines collected here should have the same
signature for the data arguments; keyword arguments can be different.
"""
from enum import Enum

__all__ = (
    'set_default_backend',
    'PlotBackend',
    'set_defaults',
    'reset_defaults',
    'show_defaults',
    'save_defaults',
    'load_defaults',
    'get_help',
    'get_plotter',
    'dos',
    'spinpol_dos',
    'bands',
    'spinpol_bands',
    'scatter',
    'line',
)

_DEFAULT_BACKEND = 'mpl'


def set_default_backend(backend):
    """
    Sets the default backend used when no explicit backend is specified.

    :param backend: Name of the backend to use
    """
    global _DEFAULT_BACKEND  #pylint: disable=global-statement
    backend = PlotBackend.from_str(backend)
    _DEFAULT_BACKEND = backend.value


class PlotBackend(Enum):
    """
    Enumeration containing the possible names for each plotting backend
    Initialize using the :py:meth:`from_str()` method

    At the moment the following are supported (case-insensitive)

    * ``matplotlib``: either `'mpl'` or `'matplotlib'`
    * ``bokeh``: `'bokeh'`
    """
    mpl = 'matplotlib'
    bokeh = 'bokeh'

    @staticmethod
    def from_str(label):
        """
        Initialize the :py:class:`PlotBackend` from a given string

        :param label: str to use to initialize the backend
                      if it is `None` the default is returned

        :returns: :py:class:`PlotBackend` instance corresponding to the label
        """
        if label is None:
            #Default backend
            return PlotBackend.default()
        elif isinstance(label, PlotBackend):
            return label
        elif label.lower() in ('mpl', 'matplotlib'):
            return PlotBackend.mpl
        elif label.lower() in ('bokeh',):
            return PlotBackend.bokeh
        else:
            raise NotImplementedError

    @staticmethod
    def default():
        """
        Return a :py:class:`PlotBackend` instance corresponding to
        the current default backend
        """
        return PlotBackend.from_str(_DEFAULT_BACKEND)


def set_defaults(backend=None, **kwargs):
    """
    Sets defaults for the plot parameters.

    :param backend: For which backend to set the parameters

    The Kwargs are used to set the parameters of the specified backend
    """
    from .plot_methods import set_mpl_plot_defaults
    from .bokeh_plots import set_bokeh_plot_defaults

    backend = PlotBackend.from_str(backend)

    if backend == PlotBackend.mpl:
        set_mpl_plot_defaults(**kwargs)
    elif backend == PlotBackend.bokeh:
        set_bokeh_plot_defaults(**kwargs)
    else:
        raise NotImplementedError


def reset_defaults(backend=None):
    """
    Reset the defaults for theplot parameters to the original state.

    :param backend: For which backend to reset the parameters
    """
    from .plot_methods import reset_mpl_plot_defaults
    from .bokeh_plots import reset_bokeh_plot_defaults

    backend = PlotBackend.from_str(backend)

    if backend == PlotBackend.mpl:
        reset_mpl_plot_defaults()
    elif backend == PlotBackend.bokeh:
        reset_bokeh_plot_defaults()
    else:
        raise NotImplementedError


def show_defaults(backend=None):
    """
    Show the current set defaults for the plot parameters.

    :param backend: For which backend to show the parameters
    """
    from .plot_methods import show_mpl_plot_defaults
    from .bokeh_plots import show_bokeh_plot_defaults

    backend = PlotBackend.from_str(backend)

    if backend == PlotBackend.mpl:
        show_mpl_plot_defaults()
    elif backend == PlotBackend.bokeh:
        show_bokeh_plot_defaults()
    else:
        raise NotImplementedError


def save_defaults(backend=None, filename='plot_defaults.json', save_complete=False):
    """
    Save the defaults for the plot parameters.

    :param backend: For which backend to save the parameters
    :param filename: str of the filename to save the defaults to
    :param save_complete: bool, if True also the hardcoded defaults are included
    """
    from .plot_methods import save_mpl_defaults
    from .bokeh_plots import save_bokeh_defaults

    backend = PlotBackend.from_str(backend)

    if backend == PlotBackend.mpl:
        save_mpl_defaults(filename=filename, save_complete=save_complete)
    elif backend == PlotBackend.bokeh:
        save_bokeh_defaults(filename=filename, save_complete=save_complete)
    else:
        raise NotImplementedError


def load_defaults(backend=None, filename='plot_defaults.json'):
    """
    Load defaults for the plot parameters from a file and set the contained defaults.

    :param backend: For which backend to save the parameters
    :param filename: str of the filename to load the defaults from
    """
    from .plot_methods import load_mpl_defaults
    from .bokeh_plots import load_bokeh_defaults

    backend = PlotBackend.from_str(backend)

    if backend == PlotBackend.mpl:
        load_mpl_defaults(filename=filename)
    elif backend == PlotBackend.bokeh:
        load_bokeh_defaults(filename=filename)
    else:
        raise NotImplementedError


def get_help(key, backend=None):
    """
    Get a help string for a given parameter.

    :param key: name of the parameter to get the parameter for
    :param backend: For which backend to get the  description of the parameter
    """
    from .plot_methods import get_mpl_help
    from .bokeh_plots import get_bokeh_help

    backend = PlotBackend.from_str(backend)

    if backend == PlotBackend.mpl:
        get_mpl_help(key)
    elif backend == PlotBackend.bokeh:
        get_bokeh_help(key)
    else:
        raise NotImplementedError


def get_plotter(backend=None):
    """
    Get the instance of the :py:class:`~masci_tools.vis.parameters.Plotter` subclass
    used for the given plotting backend

    :param backend: For which backend to get the Plotter instace
    """
    import masci_tools.vis.plot_methods as mpl
    import masci_tools.vis.bokeh_plots as bok

    backend = PlotBackend.from_str(backend)

    if backend == PlotBackend.mpl:
        return mpl.plot_params
    elif backend == PlotBackend.bokeh:
        return bok.plot_params
    else:
        raise NotImplementedError


def dos(energy_grid, dos_data, backend=None, data=None, **kwargs):
    """
    Plot the provided data as a density of states (not spin-polarized). Can be done
    horizontally or vertical via the switch `xyswitch`

    :param energy_grid: data for the energy grid of the DOS
    :param dos_data: data for all the DOS components to plot
    :param data: source for the data of the plot (optional) (pandas Dataframe for example)
    :param backend: name of the backend to use (uses a default if None is given)

    Kwargs are passed on to the backend plotting functions:

        - ``matplotlib``: :py:func:`~masci_tools.vis.plot_methods.plot_dos()`
        - ``bokeh``: :py:func:`~masci_tools.vis.bokeh_plots.bokeh_dos()`

    :returns: Figure object for the used plotting backend
    """
    from .plot_methods import plot_dos
    from .bokeh_plots import bokeh_dos

    plot_funcs = {PlotBackend.mpl: plot_dos, PlotBackend.bokeh: bokeh_dos}

    backend = PlotBackend.from_str(backend)

    return plot_funcs[backend](energy_grid, dos_data, data=data, **kwargs)


def spinpol_dos(energy_grid, dos_data_up, dos_data_dn, backend=None, data=None, **kwargs):
    """
    Plot the provided data as a density of states (spin-polarized). Can be done
    horizontally or vertical via the switch `xyswitch`

    :param energy_grid: data for the energy grid of the DOS
    :param dos_data_up: data for all the DOS components to plot for spin-up
    :param dos_data_dn: data for all the DOS components to plot for spin-down
    :param data: source for the data of the plot (optional) (pandas Dataframe for example)
    :param backend: name of the backend to use (uses a default if None is given)

    Kwargs are passed on to the backend plotting functions:

        - ``matplotlib``: :py:func:`~masci_tools.vis.plot_methods.plot_spinpol_dos()`
        - ``bokeh``: :py:func:`~masci_tools.vis.bokeh_plots.bokeh_spinpol_dos()`

    :returns: Figure object for the used plotting backend
    """
    from .plot_methods import plot_spinpol_dos
    from .bokeh_plots import bokeh_spinpol_dos

    plot_funcs = {PlotBackend.mpl: plot_spinpol_dos, PlotBackend.bokeh: bokeh_spinpol_dos}

    backend = PlotBackend.from_str(backend)

    return plot_funcs[backend](energy_grid, dos_data_up, dos_data_dn, data=data, **kwargs)


def bands(kpath, eigenvalues, backend=None, data=None, **kwargs):
    """
    Plot the provided data for a bandstructure (non spin-polarized)
    Non-weighted, weighted, as a line plot or scatter plot,
    color-mapped or fixed colors are all possible options

    :param kpath: data for the kpoints path (flattened to 1D)
    :param eigenvalues: data for the eigenvalues
    :param data: source for the data of the plot (optional) (pandas Dataframe for example)
    :param backend: name of the backend to use (uses a default if None is given)

    Kwargs are passed on to the backend plotting functions:

        - ``matplotlib``: :py:func:`~masci_tools.vis.plot_methods.plot_bands()`
        - ``bokeh``: :py:func:`~masci_tools.vis.bokeh_plots.bokeh_bands()`

    :returns: Figure object for the used plotting backend
    """
    from .plot_methods import plot_bands
    from .bokeh_plots import bokeh_bands

    plot_funcs = {PlotBackend.mpl: plot_bands, PlotBackend.bokeh: bokeh_bands}

    backend = PlotBackend.from_str(backend)

    return plot_funcs[backend](kpath, eigenvalues, data=data, **kwargs)


def spinpol_bands(kpath, eigenvalues_up, eigenvalues_dn, backend=None, data=None, **kwargs):
    """
    Plot the provided data for a bandstructure (spin-polarized)
    Non-weighted, weighted, as a line plot or scatter plot,
    color-mapped or fixed colors are all possible options

    :param kpath: data for the kpoints path (flattened to 1D)
    :param eigenvalues_up: data for the eigenvalues for spin-up
    :param eigenvalues_dn: data for the eigenvalues for spin-down
    :param data: source for the data of the plot (optional) (pandas Dataframe for example)
    :param backend: name of the backend to use (uses a default if None is given)

    Kwargs are passed on to the backend plotting functions:

        - ``matplotlib``: :py:func:`~masci_tools.vis.plot_methods.plot_spinpol_bands()`
        - ``bokeh``: :py:func:`~masci_tools.vis.bokeh_plots.bokeh_spinpol_bands()`

    :returns: Figure object for the used plotting backend
    """
    from .plot_methods import plot_spinpol_bands
    from .bokeh_plots import bokeh_spinpol_bands

    plot_funcs = {PlotBackend.mpl: plot_spinpol_bands, PlotBackend.bokeh: bokeh_spinpol_bands}

    backend = PlotBackend.from_str(backend)

    return plot_funcs[backend](kpath, eigenvalues_up, eigenvalues_dn, data=data, **kwargs)


def scatter(xdata, ydata, backend=None, data=None, **kwargs):
    """
    Plot the provided data as a scatter plot. Varying size and color
    are possible. Multiple data sets are possible

    :param xdata: data for the x-axis
    :param xdata: data for the y-axis
    :param data: source for the data of the plot (optional) (pandas Dataframe for example)
    :param backend: name of the backend to use (uses a default if None is given)

    Kwargs are passed on to the backend plotting functions:

        - ``matplotlib``: :py:func:`~masci_tools.vis.plot_methods.multi_scatter_plot()`
        - ``bokeh``: :py:func:`~masci_tools.vis.bokeh_plots.bokeh_multi_scatter()`

    :returns: Figure object for the used plotting backend
    """
    from .plot_methods import multi_scatter_plot
    from .bokeh_plots import bokeh_multi_scatter

    plot_funcs = {PlotBackend.mpl: multi_scatter_plot, PlotBackend.bokeh: bokeh_multi_scatter}

    backend = PlotBackend.from_str(backend)

    return plot_funcs[backend](xdata, ydata, data=data, **kwargs)


def line(xdata, ydata, backend=None, data=None, **kwargs):
    """
    Plot the provided data as a line plot. Multiple data sets are possible

    :param xdata: data for the x-axis
    :param xdata: data for the y-axis
    :param data: source for the data of the plot (optional) (pandas Dataframe for example)
    :param backend: name of the backend to use (uses a default if None is given)

    Kwargs are passed on to the backend plotting functions:

        - ``matplotlib``: :py:func:`~masci_tools.vis.plot_methods.multiple_scatterplots()`
        - ``bokeh``: :py:func:`~masci_tools.vis.bokeh_plots.bokeh_line()`

    :returns: Figure object for the used plotting backend
    """
    from .plot_methods import multiple_scatterplots
    from .bokeh_plots import bokeh_line

    plot_funcs = {PlotBackend.mpl: multiple_scatterplots, PlotBackend.bokeh: bokeh_line}

    backend = PlotBackend.from_str(backend)

    return plot_funcs[backend](xdata, ydata, data=data, **kwargs)



# -*- coding: utf-8 -*-
"""
Here the functions from input_transforms are imported to trigger their decorators
This is needed to make the HDF5Reader aware of their existence
"""

from .reader import HDF5Reader
from .transforms import HDF5TransformationError

__all__ = ['HDF5Reader', 'HDF5TransformationError']



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""
functions to extract information about the fleur schema input or output
"""
from masci_tools.util.case_insensitive_dict import CaseInsensitiveDict, CaseInsensitiveFrozenSet
from functools import wraps
from collections import namedtuple
import warnings
import math

#These types have infinite recursive paths and CANNOT BE PARSED in the path generation
_RECURSIVE_TYPES = ['CompositeTimerType']
#Name of the type of an scf iteration in the out schema (At this point the paths are split up)
_ITERATION_TYPE = 'IterationType'
_INPUT_TYPE = 'FleurInputType'

# The types defined here should not be reduced further and are associated with one clear base type
# AngularMomentumNumberType and MainQuantumNumberType are here because they are integers
# but are implemented as xsd:string with a regex
BASE_TYPES = {
    'switch': {'FleurBool'},
    'int': {
        'xsd:nonNegativeInteger', 'xsd:positiveInteger', 'xsd:integer', 'AngularMomentumNumberType',
        'MainQuantumNumberType'
    },
    'float': {'xsd:double'},
    'float_expression': {'FleurDouble'},
    'string': {'xsd:string'}
}

AttributeType = namedtuple('AttributeType', ('base_type', 'length'))

#We define some decorators to cache results to not repeat too many similar xpath calls or recursive function calls


def _cache_xpath_construction(func):
    """
    Decorator for the `_get_xpath` and `_get_attrib_xpath` functions to speed up the parsing of
    xml schemas by caching results
    """

    results = {}

    @wraps(func)
    def wrapper(xmlschema, namespaces, name, **kwargs):
        """
        This function produces a hash from all the arguments modifying the behaviour of the wrapped function
        and looks up results in dict based on this hash. If the version of the schema
        is different than before or the dict contains more than 1024 entries the cache is cleared
        """

        version = str(xmlschema.xpath('/xsd:schema/@version', namespaces=namespaces)[0])
        root_tag = str(xmlschema.xpath('/xsd:schema/xsd:element/@name', namespaces=namespaces)[0])

        arg_tuple = (version, root_tag, name, kwargs.get('enforce_end_type', ''), kwargs.get('ref', '')) + \
                    tuple(key for key in kwargs if kwargs.get(key, False))

        hash_args = hash(arg_tuple)
        if version not in results:
            results.clear()
            results[version] = {}

        if hash_args not in results[version]:
            res = func(xmlschema, namespaces, name, **kwargs)
            if len(results[version]) >= 1024:
                results[version].clear()
                return res
            results[version][hash_args] = res

        return results[version][hash_args].copy()

    return wrapper


def _cache_xpath_eval(func):
    """
    Decorator for the `_xpath_eval` function to speed up concrete xpath calls on the schema
    by caching the results
    """
    results = {}

    @wraps(func)
    def wrapper(xmlschema, xpath, namespaces):
        """
        This function produces a hash from all the arguments modifying the behaviour of the wrapped function
        and looks up results in dict based on this hash. If the version of the schema
        is different than before or the dict contains more than 1024 entries the cache is cleared
        """

        version = str(xmlschema.xpath('/xsd:schema/@version', namespaces=namespaces)[0])
        root_tag = str(xmlschema.xpath('/xsd:schema/xsd:element/@name', namespaces=namespaces)[0])

        arg_tuple = (version, root_tag, xpath, *namespaces.values())

        hash_args = hash(arg_tuple)
        if version not in results:
            results.clear()
            results[version] = {}

        if hash_args not in results[version]:
            res = func(xmlschema, xpath, namespaces)
            if len(results[version]) >= 1024:
                results[version].clear()
                return res
            results[version][hash_args] = res

        return results[version][hash_args].copy()

    return wrapper


@_cache_xpath_eval
def _xpath_eval(xmlschema, xpath, namespaces):
    """
    Wrapper around the xpath calls in this module. Used for caching the
    results

    :param xmlschema: xmltree representing the schema
    :param xpath: str, xpath expression to evaluate
    :param namespaces: dictionary with the defined namespaces
    """
    return xmlschema.xpath(xpath, namespaces=namespaces)


def _remove_xsd_namespace(tag, namespaces):
    """
    Strips the xsd namespace prefix from tags to make the functions more understandable

    :param tag: tag containing the xsd namespace prefix
    :param namespaces: dictionary with the defined namespaces

    :return: tag with the xsd namespace removed
    """
    return tag.replace(f"{'{'}{namespaces['xsd']}{'}'}", '')


def _is_base_type(type_name):
    """
    Return whether the given type_name appears in the sets in the BASE_TYPES dict
    """
    return any(type_name in types for types in BASE_TYPES.values())


def _get_parent_fleur_type(elem, namespaces, stop_non_unique=False):
    """
    Returns the parent simple or complexType to the given element
    If stop_sequence is given and True None is returned when a sequence is encountered
    in the parent chain

    :param elem: etree element, starting element
    :param namespaces: dictionary with the defined namespaces
    :param stop_sequence: If a sequence is encountered in the loop it alos terminates

    :return: the element of the parent type and the tag of the parent type with the namespaces removed
    """
    valid_end_types = ['simpleType', 'complexType', 'group']
    parent = elem.getparent()
    parent_type = _remove_xsd_namespace(parent.tag, namespaces)
    if stop_non_unique:
        if 'maxOccurs' in parent.attrib:
            if parent.attrib['maxOccurs'] != '1':
                return None, None
    while parent_type not in valid_end_types:
        parent = parent.getparent()
        parent_type = _remove_xsd_namespace(parent.tag, namespaces)
        if stop_non_unique:
            if 'maxOccurs' in parent.attrib:
                if parent.attrib['maxOccurs'] != '1':
                    return None, None
    return parent, parent_type


def _get_base_types(xmlschema, namespaces, type_elem, convert_to_base=True, basic_types_mapping=None):
    """
    Analyses the given type element to deduce its base_types and length restrictions

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces
    :param type_elem: etree element of the type to analyse
    :param convert_to_base: if True all possible types are converted to their base_types using either base_types
                            or basic_types_mapping
    :param basic_types_mapping: dictionary with definitions of basic_types (used for the outputschema to get the
                                definitions from the inputschema)

    :return: list of valid possibilities (AttributeType tuples) of base types and length restrictions
    """

    if basic_types_mapping is None:
        basic_types_mapping = {}

    length = _get_length(xmlschema, namespaces, type_elem)

    possible_types = set()
    for child in type_elem:
        child_type = _remove_xsd_namespace(child.tag, namespaces)

        types = None
        if child_type in ('restriction', 'extension'):
            types = [child.attrib['base']]
        elif child_type == 'list':
            types = [child.attrib['itemType']]
        elif child_type == 'union' and 'memberTypes' in child.attrib:
            types = child.attrib['memberTypes'].split(' ')
        elif child_type in ('union', 'simpleType'):
            new_types = _get_base_types(xmlschema,
                                        namespaces,
                                        child,
                                        convert_to_base=False,
                                        basic_types_mapping=basic_types_mapping)
            possible_types.update(new_types)

        #Reduce the found types to types that can be converted to base types
        if types is not None:

            for found_type in types:

                if _is_base_type(found_type):
                    possible_types.add((found_type, length))
                elif found_type in basic_types_mapping:
                    possible_types.add((found_type, length))
                else:
                    sub_types = _xpath_eval(xmlschema, f"//xsd:simpleType[@name='{found_type}']", namespaces=namespaces)
                    if len(sub_types) == 0:
                        sub_types = _xpath_eval(xmlschema,
                                                f"//xsd:complexType[@name='{found_type}']/xsd:simpleContent",
                                                namespaces=namespaces)

                    if len(sub_types) == 0:
                        raise ValueError(f"No such type '{found_type}'")

                    if len(sub_types) > 1:
                        raise ValueError(f"No unique type found for '{found_type}'")

                    new_types = _get_base_types(xmlschema,
                                                namespaces,
                                                sub_types[0],
                                                convert_to_base=False,
                                                basic_types_mapping=basic_types_mapping)

                    if length != 1:
                        possible_types.update((base_type, length) for base_type, _ in new_types)
                    else:
                        possible_types.update(new_types)

    if convert_to_base:
        converted_types = set()
        for found_type, length in possible_types:
            if found_type in basic_types_mapping:
                converted_types.update(basic_types_mapping[found_type])
            else:
                for base_type, possible_base_types in BASE_TYPES.items():
                    if found_type in possible_base_types:
                        converted_types.add(AttributeType(base_type=base_type, length=length))
        possible_types = sorted(converted_types, key=type_order)

        if any(type_def.length is None for type_def in possible_types):
            raise ValueError(f'Length determination failed {type_elem}: {possible_types}')

    return possible_types


def _get_length(xmlschema, namespaces, type_elem):
    """
    Analyse the given type to determine, whether there is a length restriction

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces
    :param type_elem: etree.Element of the type to analyse

    :return: if a length restriction is found return the value,
             if a list with no restriction is found return 'unbounded',
             if neither are found return 1
    """

    type_tag = _remove_xsd_namespace(type_elem.tag, namespaces)

    if type_tag == 'simpleType':

        child = type_elem.getchildren()
        if len(child) != 1:
            return 1
        child = child[0]

        child_type = _remove_xsd_namespace(child.tag, namespaces)
        if child_type == 'restriction':
            for restriction_child in child:
                restr_type = _remove_xsd_namespace(restriction_child.tag, namespaces)
                if restr_type == 'length':
                    return int(restriction_child.attrib['value'])
        elif child_type == 'list':
            return 'unbounded'

        return 1

    elif type_tag in ('complexType', 'simpleContent'):
        if 'name' in type_elem.attrib:
            type_name = type_elem.attrib['name']
        else:
            type_name = type_elem.getparent().attrib['name']

        base_type = _xpath_eval(xmlschema,
                                f"//xsd:complexType[@name='{type_name}']/xsd:simpleContent/xsd:extension/@base",
                                namespaces=namespaces)
        if len(base_type) == 0:
            return 1

        base_type_elem = _xpath_eval(xmlschema, f"//xsd:simpleType[@name='{base_type[0]}']", namespaces=namespaces)

        if len(base_type_elem) == 0:
            return 1

        return _get_length(xmlschema, namespaces, base_type_elem[0])

    else:
        return None


@_cache_xpath_construction
def _get_xpath(xmlschema,
               namespaces,
               tag_name,
               enforce_end_type=None,
               ref=None,
               stop_non_unique=False,
               stop_iteration=False,
               iteration_root=False):
    """
    construct all possible simple xpaths to a given tag

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces
    :param tag_name: name of the starting tag
    :param enforce_end_type: If given the type of the starting tag has o match this string
    :param ref: If given we start from a group reference with this name
    :param stop_non_unique: If True all paths, where one tag has maxOccurs!=1 is discarded
    :param stop_iteration: If True the path generation discards all paths going through a iteration element
    :param iteration_root: If True the path generation only generates paths going through a iteration but
                       terminates as if the group element is the root of the file
    :return: None if no path is found, if a single path is found return the string of the path,
             otherwise a list with all possible paths is returned
    """

    possible_paths = set()
    if enforce_end_type in _RECURSIVE_TYPES:
        return possible_paths
    root_tag = get_root_tag(xmlschema, namespaces)
    if tag_name == root_tag:
        if not iteration_root:
            possible_paths.add(f'/{root_tag}')
        return possible_paths

    #Get all possible starting points
    if ref is not None:
        startPoints = _xpath_eval(xmlschema, f"//xsd:group[@ref='{ref}']", namespaces=namespaces)
    else:
        if enforce_end_type is None:
            startPoints = _xpath_eval(xmlschema, f"//xsd:element[@name='{tag_name}']", namespaces=namespaces)
        else:
            startPoints = _xpath_eval(xmlschema,
                                      f"//xsd:element[@name='{tag_name}' and @type='{enforce_end_type}']",
                                      namespaces=namespaces)
    if stop_non_unique:
        startPoints_copy = startPoints.copy()
        for point in startPoints_copy:
            if 'maxOccurs' in point.attrib:
                if point.attrib['maxOccurs'] != '1':
                    startPoints.remove(point)
    for elem in startPoints:
        currentelem = elem
        currentTag = tag_name
        parent_type, parent_tag = _get_parent_fleur_type(currentelem, namespaces, stop_non_unique=stop_non_unique)
        if parent_type is None:
            continue
        next_type = parent_type.attrib['name']

        if next_type == _ITERATION_TYPE:
            if stop_iteration:
                continue
            if iteration_root:
                possible_paths.add(f'./{currentTag}')
                return possible_paths

        if parent_tag == 'group':
            possible_paths_group = _get_xpath(xmlschema,
                                              namespaces,
                                              currentTag,
                                              ref=next_type,
                                              stop_non_unique=stop_non_unique,
                                              stop_iteration=stop_iteration,
                                              iteration_root=iteration_root)
            for grouppath in possible_paths_group:
                possible_paths.add(grouppath)
        else:
            if stop_non_unique:
                currentelem = _xpath_eval(
                    xmlschema,
                    f"//xsd:element[@type='{next_type}' and @maxOccurs=1] | //xsd:element[@type='{next_type}' and not(@maxOccurs)]",
                    namespaces=namespaces)
            else:
                currentelem = _xpath_eval(xmlschema, f"//xsd:element[@type='{next_type}']", namespaces=namespaces)

            if len(currentelem) == 0:
                continue
            for new_elem in currentelem:
                newTag = new_elem.attrib['name']
                possible_paths_tag = _get_xpath(xmlschema,
                                                namespaces,
                                                newTag,
                                                enforce_end_type=next_type,
                                                stop_non_unique=stop_non_unique,
                                                stop_iteration=stop_iteration,
                                                iteration_root=iteration_root)
                for tagpath in possible_paths_tag:
                    possible_paths.add(f'{tagpath}/{tag_name}')

    if iteration_root:
        #Remove any path that slipped through and contains the root tag of the out file
        possible_paths_copy = possible_paths.copy()
        for path in possible_paths_copy:
            if root_tag in path:
                possible_paths.discard(path)

    return possible_paths


def _get_contained_attribs(xmlschema, namespaces, elem, optional=False):
    """
    Get all defined attributes contained in the given etree Element of the schema

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces
    :param elem: etree Element to analyse
    :param optional: bool, if True only the attributes which have use='optional'
                     are added and their default value (defaults to None) is returned in a
                     :py:class:`masci_tools.util.case_insensitive_dict.CaseInsensitiveDict`

    :raises: AssertionError if case insensitivity lead to lost information

    :returns: CaseInsensitiveDict (optional=True) or CaseInsensitiveFrozenSet with
              the attribute names
    """
    attrib_list = []
    for child in elem:
        child_type = _remove_xsd_namespace(child.tag, namespaces)

        if child_type == 'attribute':
            if optional and child.attrib.get('use', 'required') == 'optional':
                name = child.attrib['name']
                default = child.attrib.get('default', None)
                attrib_list.append((name, default))
            elif not optional:
                attrib_list.append(child.attrib['name'])
        elif child_type in ['simpleContent', 'extension']:
            new_attribs = _get_contained_attribs(xmlschema, namespaces, child, optional=optional)
            if optional:
                for entry in new_attribs.items():
                    attrib_list.append(entry)
            else:
                for attrib in new_attribs:
                    attrib_list.append(new_attribs.original_case[attrib])

    if not optional:
        attrib_res = CaseInsensitiveFrozenSet(attrib_list)
        assert len(set(attrib_list)) == len(attrib_res), f'Lost Information: {attrib_list}'
    else:
        attrib_res = CaseInsensitiveDict(attrib_list)

    return attrib_res


def _get_optional_tags(xmlschema, namespaces, elem):
    """
    Get all defined tags contained in the given etree Element of the schema
    with minOccurs=0

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces
    :param elem: etree Element to analyse

    :raises: AssertionError if case insensitivity lead to lost information

    :returns: CaseInsensitiveFrozenSet with the tag names
    """
    optional_list = []
    for child in elem:
        child_type = _remove_xsd_namespace(child.tag, namespaces)

        if child_type == 'element':
            if 'minOccurs' in child.attrib:
                if child.attrib['minOccurs'] == '0':
                    optional_list.append(child.attrib['name'])
        elif child_type in ['sequence', 'all', 'choice']:
            new_optionals = _get_optional_tags(xmlschema, namespaces, child)
            for opt in new_optionals:
                optional_list.append(new_optionals.original_case[opt])

    optional_set = CaseInsensitiveFrozenSet(optional_list)
    assert len(set(optional_list)) == len(optional_set), f'Lost Information: {optional_list}'

    return optional_set


def _is_simple(namespaces, elem):
    """
    Determine if a given etree element is simple (only contains attributes or text (no sub elements))

    :param namespaces: dictionary with the defined namespaces
    :param elem: etree Element to analyse

    :raises: ValueError if an unknown type is encountered

    :returns: bool determining, whether the element is simple
    """
    simple = True
    for child in elem:
        child_type = _remove_xsd_namespace(child.tag, namespaces)

        if child_type in ['attribute', 'simpleContent']:
            continue
        if child_type in ['element', 'sequence', 'choice', 'all']:
            simple = False
        else:
            raise ValueError(f"Don't know what to do with '{child_type}'")

    return simple


def _get_simple_tags(xmlschema, namespaces, elem, input_mapping=None):
    """
    Get all defined tags contained in the given etree Element of the schema
    which can only contain attributes or text (no sub elements)

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces
    :param elem: etree Element to analyse
    :param input_mapping: dict, with the defined types from the input schema

    :raises: AssertionError if case insensitivity lead to lost information

    :returns: CaseInsensitiveFrozenSet with the tag names
    """

    if input_mapping is None:
        input_mapping = {}

    simple_list = []
    for child in elem:
        child_type = _remove_xsd_namespace(child.tag, namespaces)

        if child_type == 'element':
            if child.attrib['type'] == _INPUT_TYPE:
                continue
            if child.attrib['type'] in input_mapping:
                simple_list.append(child.attrib['name'])
                continue

            type_elem = _xpath_eval(xmlschema,
                                    f"//xsd:simpleType[@name='{child.attrib['type']}']",
                                    namespaces=namespaces)
            if len(type_elem) != 0:
                simple_list.append(child.attrib['name'])
            else:
                type_elem = _xpath_eval(xmlschema,
                                        f"//xsd:complexType[@name='{child.attrib['type']}']",
                                        namespaces=namespaces)
                if len(type_elem) == 0:
                    simple_list.append(child.attrib['name'])
                elif _is_simple(namespaces, type_elem[0]):
                    simple_list.append(child.attrib['name'])
        elif child_type in ['sequence', 'all', 'choice']:
            new_simple = _get_simple_tags(xmlschema, namespaces, child, input_mapping=input_mapping)
            for simple in new_simple:
                simple_list.append(new_simple.original_case[simple])

    simple_set = CaseInsensitiveFrozenSet(simple_list)
    assert len(set(simple_list)) == len(simple_set), f'Lost Information: {simple_list}'

    return simple_set


def _get_several_tags(xmlschema, namespaces, elem):
    """
    Get all defined tags contained in the given etree Element of the schema
    which can occur multiple times (maxOccurs!=1)

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces
    :param elem: etree Element to analyse

    :raises: AssertionError if case insensitivity lead to lost information

    :returns: CaseInsensitiveFrozenSet with the tag names
    """
    several_list = []
    for child in elem:
        child_type = _remove_xsd_namespace(child.tag, namespaces)

        if child_type == 'element':
            if 'maxOccurs' in child.attrib:
                if child.attrib['maxOccurs'] != '1':
                    several_list.append(child.attrib['name'])
        elif child_type in ['sequence', 'all', 'choice']:
            if 'maxOccurs' in child.attrib:
                if child.attrib['maxOccurs'] != '1':
                    new_several = _get_sequence_order(xmlschema, namespaces, child)
                    for tag in new_several:
                        several_list.append(tag)
            else:
                new_several = _get_several_tags(xmlschema, namespaces, child)
                for tag in new_several:
                    several_list.append(new_several.original_case[tag])

    several_set = CaseInsensitiveFrozenSet(several_list)
    assert len(set(several_list)) == len(several_set), f'Lost Information: {several_list}'

    return several_set


def _get_contained_text_tags(xmlschema, namespaces, elem, text_tags):
    """
    Get all defined tags contained in the given etree Element of the schema
    which can contain text

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces
    :param elem: etree Element to analyse
    :param text_tags: set with all known types of test elements

    :raises: AssertionError if case insensitivity lead to lost information

    :returns: CaseInsensitiveFrozenSet with the tag names
    """
    text_list = []
    for child in elem:
        child_type = _remove_xsd_namespace(child.tag, namespaces)

        if child_type == 'element':
            if child.attrib['name'] in text_tags:
                text_list.append(child.attrib['name'])
        elif child_type in ['sequence', 'all', 'choice']:
            new_tags = _get_contained_text_tags(xmlschema, namespaces, child, text_tags)
            for tag in new_tags:
                text_list.append(new_tags.original_case[tag])

    text_set = CaseInsensitiveFrozenSet(text_list)
    assert len(set(text_list)) == len(text_set), f'Lost Information: {text_list}'

    return text_set


@_cache_xpath_construction
def _get_attrib_xpath(xmlschema,
                      namespaces,
                      attrib_name,
                      stop_non_unique=False,
                      stop_iteration=False,
                      iteration_root=False):
    """
    construct all possible simple xpaths to a given attribute

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces
    :param attrib_name: name of the attribute
    :param stop_non_unique: If True all paths, where one tag has maxOccurs!=1 is discarded
    :param stop_iteration: If True the path generation discards all paths going through a iteration element
    :param iteration_root: If True the path generation only generates paths going through a iteration but
                       terminates as if the iteration element is the root of the file
    :return: None if no path is found, if a single path is found return the string of the path,
             otherwise a list with all possible paths is returned
    """
    possible_paths = set()
    attribute_tags = _xpath_eval(xmlschema, f"//xsd:attribute[@name='{attrib_name}']", namespaces=namespaces)
    for attrib in attribute_tags:
        parent_type, parent_tag = _get_parent_fleur_type(attrib, namespaces, stop_non_unique=stop_non_unique)
        if parent_type is None:
            continue

        start_type = parent_type.attrib['name']
        if start_type == _ITERATION_TYPE:
            if stop_iteration:
                continue
            if iteration_root:
                possible_paths.add(f'./@{attrib_name}')
                continue

        if stop_non_unique:
            element_tags = _xpath_eval(
                xmlschema,
                f"//xsd:element[@type='{start_type}' and @maxOccurs=1]/@name | //xsd:element[@type='{start_type}' and not(@maxOccurs)]/@name",
                namespaces=namespaces)
        else:
            element_tags = _xpath_eval(xmlschema, f"//xsd:element[@type='{start_type}']/@name", namespaces=namespaces)

        for tag in element_tags:
            tag_paths = _get_xpath(xmlschema,
                                   namespaces,
                                   tag,
                                   enforce_end_type=start_type,
                                   stop_non_unique=stop_non_unique,
                                   stop_iteration=stop_iteration,
                                   iteration_root=iteration_root)
            for path in tag_paths:
                possible_paths.add(f'{path}/@{attrib_name}')

    return possible_paths


def _get_sequence_order(xmlschema, namespaces, sequence_elem):
    """
    Extract the enforced order of elements in the given sequence element

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces
    :param sequence_elem: element of the sequence to analyse

    :return: list of tags, in the order they have to occur in
    """
    elem_order = []
    for child in sequence_elem:
        child_type = _remove_xsd_namespace(child.tag, namespaces)

        if child_type == 'element':
            elem_order.append(child.attrib['name'])
        elif child_type in ['choice', 'sequence']:
            new_order = _get_sequence_order(xmlschema, namespaces, child)
            for elem in new_order:
                elem_order.append(elem)
        elif child_type == 'group':
            group = _xpath_eval(xmlschema,
                                f"//xsd:group[@name='{child.attrib['ref']}']/xsd:sequence",
                                namespaces=namespaces)
            new_order = _get_sequence_order(xmlschema, namespaces, group[0])
            for elem in new_order:
                elem_order.append(elem)
        elif child_type in ['attribute', 'simpleContent', 'all']:
            continue
        else:
            raise KeyError(f'Dont know what to do with {child_type}')

    return elem_order


def _get_valid_tags(xmlschema, namespaces, sequence_elem):
    """
    Extract all allowed elements in the given sequence element

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces
    :param sequence_elem: element of the sequence to analyse

    :return: list of tags, in the order they have to occur in
    """
    elems = []
    for child in sequence_elem:
        child_type = _remove_xsd_namespace(child.tag, namespaces)

        if child_type == 'element':
            elems.append(child.attrib['name'])
        elif child_type in ['choice', 'sequence', 'all']:
            new_elems = _get_valid_tags(xmlschema, namespaces, child)
            for elem in new_elems:
                elems.append(elem)
        elif child_type == 'group':
            group = _xpath_eval(xmlschema,
                                f"//xsd:group[@name='{child.attrib['ref']}']/xsd:sequence",
                                namespaces=namespaces)
            new_elems = _get_valid_tags(xmlschema, namespaces, group[0])
            for elem in new_elems:
                elems.append(elem)
        elif child_type in ['attribute', 'simpleContent']:
            continue
        else:
            raise KeyError(f'Dont know what to do with {child_type}')

    return elems


def _extract_all_types(xmlschema, namespaces, elems, ignore_unknown=False, **kwargs):
    """
    Determine the required type of all given attributes/elements

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces
    :param ignore_unknown: bool, if True and a type cannot be traced back to a base type
                           nothing is done, otherwise a warning is issued

    :return: possible types of the attributes in a dictionary, if multiple
             types are possible a list is inserted for the tag
    """

    types_dict = CaseInsensitiveDict()
    for elem in elems:
        name = elem.attrib['name']
        type_name = elem.attrib['type']

        possible_types = set()

        if _is_base_type(type_name):
            for base_type, types in BASE_TYPES.items():
                if type_name in types:
                    possible_types.add(AttributeType(base_type=base_type, length=1))
        else:
            if type_name in kwargs['_basic_types']:
                possible_types.update(kwargs['_basic_types'][type_name])
            elif not ignore_unknown:
                warnings.warn(f'Unsorted type:{type_name}')
            else:
                continue

        if name in types_dict:
            types_dict[name] = types_dict[name] | possible_types
        else:
            types_dict[name] = possible_types

    return types_dict


def type_order(type_def):
    """
    Key function for sorting the type definitions to avoid conflicts

    Sorted by base_type first (bool before int, string at the end)
    and then by length in ascending order (unbounded last)

    :param type_def: definition to be sorted
    """

    if not isinstance(type_def, AttributeType):
        raise ValueError('Wrong type for type_def')

    BASE_TYPE_ORDER = ('switch', 'int', 'float', 'float_expression', 'string')

    type_index = BASE_TYPE_ORDER.index(type_def.base_type)
    length = type_def.length if type_def.length != 'unbounded' else math.inf

    return type_index, length


def extract_attribute_types(xmlschema, namespaces, **kwargs):
    """
    Determine the required type of all attributes

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces

    :return: possible types of the attributes in a dictionary, if multiple
             types are possible a list is inserted for the tag
    """
    possible_attrib = _xpath_eval(xmlschema, '//xsd:attribute', namespaces=namespaces)

    types_dict = _extract_all_types(xmlschema, namespaces, possible_attrib, **kwargs)

    for name, types in types_dict.items():
        types_dict[name] = sorted(types, key=type_order)

    return types_dict


def extract_text_types(xmlschema, namespaces, **kwargs):
    """
    Determine the required type of all elements with text

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces

    :return: possible types of the attributes in a dictionary, if multiple
             types are possible a list is inserted for the tag
    """
    possible_elems = _xpath_eval(xmlschema, '//xsd:element', namespaces=namespaces)

    types_dict = _extract_all_types(xmlschema, namespaces, possible_elems, ignore_unknown=True, **kwargs)

    for name, types in types_dict.items():
        types_dict[name] = sorted(types, key=type_order)

    return types_dict


def get_tag_paths(xmlschema, namespaces, **kwargs):
    """
    Determine simple xpaths to all possible tags

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces

    :return: possible paths of all tags in a dictionary, if multiple
             paths are possible a list is inserted for the tag
    """

    stop_iteration = kwargs.get('stop_iteration', False)
    iteration_root = kwargs.get('iteration_root', False)

    possible_tags = set(_xpath_eval(xmlschema, '//xsd:element/@name', namespaces=namespaces))
    tag_paths = CaseInsensitiveDict()
    for tag in sorted(possible_tags):
        paths = _get_xpath(xmlschema, namespaces, tag, stop_iteration=stop_iteration, iteration_root=iteration_root)
        if len(paths) == 1:
            tag_paths[tag] = paths.pop()
        elif len(paths) != 0:
            tag_paths[tag] = sorted(paths)
    return tag_paths


def get_unique_attribs(xmlschema, namespaces, **kwargs):
    """
    Determine all attributes, which can be set through set_inpchanges in aiida_fleur
    Meaning ONE possible path and no tags in the path with maxOccurs!=1

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces

    :return: dictionary with all settable attributes and the corresponding path to the tag
    """

    stop_iteration = kwargs.get('stop_iteration', False)
    iteration_root = kwargs.get('iteration_root', False)

    settable = CaseInsensitiveDict()
    possible_attrib = set(_xpath_eval(xmlschema, '//xsd:attribute/@name', namespaces=namespaces))
    for attrib in sorted(possible_attrib):
        path = _get_attrib_xpath(xmlschema,
                                 namespaces,
                                 attrib,
                                 stop_non_unique=True,
                                 stop_iteration=stop_iteration,
                                 iteration_root=iteration_root)
        if len(path) == 1:
            if attrib in settable:
                settable.pop(attrib)
            else:
                settable[attrib] = path.pop()

    for attrib in sorted(kwargs['text_tags'].original_case.values()):
        path = _get_xpath(xmlschema,
                          namespaces,
                          attrib,
                          stop_non_unique=True,
                          stop_iteration=stop_iteration,
                          iteration_root=iteration_root)

        if len(path) == 1:
            if attrib in settable:
                settable.pop(attrib)
            else:
                settable[attrib] = path.pop()

    return settable


def get_unique_path_attribs(xmlschema, namespaces, **kwargs):
    """
    Determine all attributes, with multiple possible path that do have at
    least one path with all contained tags maxOccurs!=1

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces

    :return: dictionary with all attributes and the corresponding list of paths to the tag
    """

    stop_iteration = kwargs.get('stop_iteration', False)
    iteration_root = kwargs.get('iteration_root', False)
    iteration = kwargs.get('iteration', False)

    if iteration:
        settable_key = 'iteration_unique_attribs'
        settable_contains_key = 'iteration_unique_path_attribs'
    else:
        settable_key = 'unique_attribs'
        settable_contains_key = 'unique_path_attribs'

    settable = CaseInsensitiveDict()
    possible_attrib = set(_xpath_eval(xmlschema, '//xsd:attribute/@name', namespaces=namespaces))
    for attrib in sorted(possible_attrib):
        if attrib in kwargs[settable_key]:
            continue
        path = _get_attrib_xpath(xmlschema,
                                 namespaces,
                                 attrib,
                                 stop_non_unique=True,
                                 stop_iteration=stop_iteration,
                                 iteration_root=iteration_root)
        if len(path) != 0:
            settable[attrib] = sorted(set(settable.get(attrib, [])).union(path))

    for attrib in sorted(kwargs['text_tags'].original_case.values()):
        if attrib in kwargs[settable_key]:
            continue
        path = _get_xpath(xmlschema,
                          namespaces,
                          attrib,
                          stop_non_unique=True,
                          stop_iteration=stop_iteration,
                          iteration_root=iteration_root)
        if len(path) != 0:
            settable[attrib] = sorted(set(settable.get(attrib, [])).union(path))

    return settable


def get_other_attribs(xmlschema, namespaces, **kwargs):
    """
    Determine all other attributes not contained in settable or settable_contains

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces

    :return: dictionary with all attributes and the corresponding list of paths to the tag
    """

    stop_iteration = kwargs.get('stop_iteration', False)
    iteration_root = kwargs.get('iteration_root', False)
    iteration = kwargs.get('iteration', False)

    if iteration:
        settable_key = 'iteration_unique_attribs'
        settable_contains_key = 'iteration_unique_path_attribs'
    else:
        settable_key = 'unique_attribs'
        settable_contains_key = 'unique_path_attribs'

    other = CaseInsensitiveDict()
    possible_attrib = set(_xpath_eval(xmlschema, '//xsd:attribute/@name', namespaces=namespaces))
    for attrib in sorted(possible_attrib):
        path = _get_attrib_xpath(xmlschema,
                                 namespaces,
                                 attrib,
                                 stop_iteration=stop_iteration,
                                 iteration_root=iteration_root)
        if len(path) != 0:
            if attrib in kwargs[settable_key]:
                path.discard(kwargs[settable_key][attrib])
            if attrib in kwargs[settable_contains_key]:
                for contains_path in kwargs[settable_contains_key][attrib]:
                    path.discard(contains_path)

            if len(path) != 0:
                other[attrib] = sorted(set(other.get(attrib, [])).union(path))

    for attrib in sorted(kwargs['text_tags'].original_case.values()):
        path = _get_xpath(xmlschema, namespaces, attrib, stop_iteration=stop_iteration, iteration_root=iteration_root)
        if len(path) != 0:

            if attrib in kwargs[settable_key]:
                path.discard(kwargs[settable_key][attrib])
            if attrib in kwargs[settable_contains_key]:
                for contains_path in kwargs[settable_contains_key][attrib]:
                    path.discard(contains_path)

            if len(path) != 0:
                other[attrib] = sorted(set(other.get(attrib, [])).union(path))

    return other


def get_omittable_tags(xmlschema, namespaces, **kwargs):
    """
    find tags with no attributes and, which are only used to mask a list of one other possible tag (e.g. atomSpecies)

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces

    :return: list of tags, containing only a sequence of one allowed tag
    """

    possible_tags = _xpath_eval(xmlschema, '//xsd:element', namespaces=namespaces)

    omittable_tags = []
    for tag in possible_tags:
        tag_type = tag.attrib['type']
        tag_name = tag.attrib['name']

        if tag_name not in omittable_tags:
            type_elem = _xpath_eval(xmlschema, f"//xsd:complexType[@name='{tag_type}']", namespaces=namespaces)
            if len(type_elem) == 0:
                continue
            type_elem = type_elem[0]

            omittable = False
            for child in type_elem:
                child_type = _remove_xsd_namespace(child.tag, namespaces)

                if child_type == 'sequence':
                    allowed_tags = 0
                    for sequence_elem in child:
                        elem_type = _remove_xsd_namespace(sequence_elem.tag, namespaces)
                        if elem_type == 'element':
                            allowed_tags += 1
                        else:
                            allowed_tags = 0  #So that it is not unintentionally accepted
                            break
                    if allowed_tags == 1:
                        omittable = True
                else:
                    omittable = False

            if omittable:
                omittable_tags.append(tag_name)
    return omittable_tags


def get_text_tags(xmlschema, namespaces, **kwargs):
    """
    find all elements, who can contain text

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces

    :return: dictionary with tags and their corresponding type_definition
             meaning a dicationary with possible base types and evtl. length restriction
    """

    elements = _xpath_eval(xmlschema, '//xsd:element', namespaces=namespaces)

    text_tag_list = []
    for elem in elements:
        name_elem = elem.attrib['name']
        type_name = elem.attrib['type']

        if _is_base_type(type_name) or type_name in kwargs['_basic_types']:
            text_tag_list.append(name_elem)
        else:
            continue  #This type cannot be traced back to a basic type

    text_tags = CaseInsensitiveFrozenSet(text_tag_list)
    assert len(set(text_tag_list)) == len(text_tags), f'Lost Information: {text_tag_list}'

    return text_tags


def get_basic_types(xmlschema, namespaces, **kwargs):
    """
    find all types, which can be traced back directly to a base_type

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces

    :return: dictionary with type names and their corresponding type_definition
             meaning a dicationary with possible base types and evtl. length restriction
    """
    basic_type_elems = _xpath_eval(xmlschema, '//xsd:simpleType[@name]', namespaces=namespaces)
    complex_type_elems = _xpath_eval(xmlschema, '//xsd:complexType/xsd:simpleContent', namespaces=namespaces)

    basic_types = {}
    for type_elem in basic_type_elems + complex_type_elems:
        if 'name' in type_elem.attrib:
            type_name = type_elem.attrib['name']
        else:
            type_name = type_elem.getparent().attrib['name']

        if _is_base_type(type_name):
            continue  #Already a base type

        types = _get_base_types(xmlschema, namespaces, type_elem, basic_types_mapping=kwargs.get('input_basic_types'))

        if type_name not in basic_types:
            basic_types[type_name] = types
        else:
            raise ValueError(f'Already defined type {type_name}')

    #Append the definitions form the inputschema since including it directly is very messy
    if 'input_basic_types' in kwargs:
        if any(key in basic_types for key in kwargs['input_basic_types']):
            raise ValueError('Doubled type definitions from Inputschema')
        basic_types.update(kwargs['input_basic_types'])

    return basic_types


def get_tag_info(xmlschema, namespaces, **kwargs):
    """
    Get all important information about the tags
        - allowed attributes
        - contained tags (simple (only attributes), optional (with default values), several, order, text tags)

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces

    :return: dictionary with the tag information
    """

    stop_iteration = kwargs.get('stop_iteration', False)
    iteration_root = kwargs.get('iteration_root', False)

    tag_info = {}

    possible_tags = _xpath_eval(xmlschema, '//xsd:element', namespaces=namespaces)

    for tag in possible_tags:

        name_tag = tag.attrib['name']
        type_tag = tag.attrib['type']

        #Get the xpath for this tag
        tag_path = _get_xpath(xmlschema,
                              namespaces,
                              name_tag,
                              enforce_end_type=type_tag,
                              stop_iteration=stop_iteration,
                              iteration_root=iteration_root)

        tag_path = list(tag_path)

        type_elem = _xpath_eval(xmlschema, f"//xsd:complexType[@name='{type_tag}']", namespaces=namespaces)
        if len(type_elem) == 0:
            continue
        type_elem = type_elem[0]

        info_dict = {}
        info_dict['attribs'] = _get_contained_attribs(xmlschema, namespaces, type_elem)
        info_dict['optional_attribs'] = _get_contained_attribs(xmlschema, namespaces, type_elem, optional=True)
        info_dict['optional'] = _get_optional_tags(xmlschema, namespaces, type_elem)
        info_dict['several'] = _get_several_tags(xmlschema, namespaces, type_elem)
        info_dict['order'] = _get_sequence_order(xmlschema, namespaces, type_elem)
        info_dict['simple'] = _get_simple_tags(xmlschema,
                                               namespaces,
                                               type_elem,
                                               input_mapping=kwargs.get('_input_basic_types'))
        valid_tags = _get_valid_tags(xmlschema, namespaces, type_elem)
        info_dict['complex'] = CaseInsensitiveFrozenSet(valid_tags).difference(info_dict['simple'])
        info_dict['text'] = _get_contained_text_tags(xmlschema, namespaces, type_elem, kwargs['text_tags'])

        if any(len(elem) != 0 for elem in info_dict.values()):
            for path in tag_path:
                tag_info[path] = info_dict

    return tag_info


def get_root_tag(xmlschema, namespaces, **kwargs):
    """
    Returns the tag for the root element of the xmlschema

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces

    :return: name of the single element defined in the first level of the schema
    """
    return str(_xpath_eval(xmlschema, '/xsd:schema/xsd:element/@name', namespaces=namespaces)[0])


def get_input_tag(xmlschema, namespaces, **kwargs):
    """
    Returns the tag for the input type element of the outxmlschema

    :param xmlschema: xmltree representing the schema
    :param namespaces: dictionary with the defined namespaces

    :return: name of the element with the type 'FleurInputType'
    """
    return str(_xpath_eval(xmlschema, f"//xsd:element[@type='{_INPUT_TYPE}']/@name", namespaces=namespaces)[0])



# -*- coding: utf-8 -*-
###############################################################################
# Copyright (c), Forschungszentrum Jülich GmbH, IAS-1/PGI-1, Germany.         #
#                All rights reserved.                                         #
# This file is part of the Masci-tools package.                               #
# (Material science tools)                                                    #
#                                                                             #
# The code is hosted on GitHub at https://github.com/judftteam/masci-tools.   #
# For further information on the license, see the LICENSE.txt file.           #
# For further information please visit http://judft.de/.                      #
#                                                                             #
###############################################################################
"""This file contains a class to compute the crystalfield
   coefficients from convoluting the charge density with the potential
   which produces the crystalfield. This is both compatible with
   the Yttrium-Analogue approximation and self-consitent calculation of
   the potential

"""
import h5py
import os
import csv
from collections import namedtuple, defaultdict
import numpy as np
import matplotlib.pyplot as plt
import matplotlib as mpl
from pathlib import Path
from scipy.interpolate import interp1d
from scipy.special import sph_harm  #pylint: disable=no-name-in-module
from masci_tools.util.constants import HTR_TO_KELVIN
from masci_tools.io.common_functions import skipHeader

#This namedtuple is used as the return value for the crystal field calculation to have easy access
#to all the necessary information
CFCoefficient = namedtuple('CFCoefficient', ['l', 'm', 'spin_up', 'spin_down', 'unit', 'convention'])


class CFCalculation:
    r"""Class for calculating Crystal Field coefficients using the procedure
    described in C.E. Patrick, J.B. Staunton: J. Phys.: Condens. Matter 31, 305901 (2019)

    Using the formula:
        .. math::
            B_{lm} = \sqrt{\frac{2l+1}{4\pi}} \int^{R_{MT}}\! dr r^2 V_{lm}(r)n_{4f}(r)

        The read in quantities are interpolated from logarithmic meshes to equidistant meshes

        The function constructs an equidistant mesh between 0 and the muffin tin radius
        defined in `self.reference_radius` and with `self.radial_points` points

    Parameters:
        :param radial_points: int, number of radial points in the interpolated mesh
        :param reference_radius: stror float; Either 'pot' or 'cdn' or explicit number. Defines which muffin-tin radius
                                 is used for the equidistant mesh.
                                 IMPORTANT! If txt files are used the muffin-tin radius has to be provided explicitly
        :param pot_cutoff: float Defines minimum value that has to appear in potentials to not be omitted (Only HDF)
        :param only_m0: bool, Ignores coefficients with m!=0 if True
        :param quiet: bool, supresses print statements if True

   """

    __version__ = '0.1.0'

    #prefactor for converting Blm to Alm<r^l>
    _alphalm = {
        (0, 0): 1.0,
        (2, 0): 0.5,
        (2, 2): np.sqrt(6.0) / 2.0,
        (4, 0): 1.0 / 8.0,
        (4, 2): np.sqrt(10.0) / 4.0,
        (4, 3): -np.sqrt(35.0) / 2.0,
        (4, 4): np.sqrt(70.0) / 8.0,
        (6, 0): 1.0 / 16.0,
        (6, 2): np.sqrt(105.0) / 16.0,
        (6, 3): -np.sqrt(105.0) / 8.0,
        (6, 4): 3.0 * np.sqrt(14.0) / 16.0,
        (6, 6): np.sqrt(231.0) / 16.0,
    }

    def __init__(self, radial_points=4000, reference_radius='pot', pot_cutoff=1e-3, only_m0=False, quiet=False):

        self.vlm = {}
        self.cdn = {}

        self.denNorm = None
        self.phi = None
        self.theta = None

        self.interpolated = False
        self.int = {}
        self.bravaisMat = {}

        self.radial_points = radial_points
        self.reference_radius = reference_radius
        self.pot_cutoff = pot_cutoff
        self.only_m0 = only_m0
        self.quiet = quiet

    def prefactor(self, l, m):
        """Gives the lm dependent prefactor for conversion between
        Blm and Alm coefficients

        Args:
            :param l: int; orbital quantum number
            :param m: int; magnetic quantum number

        :returns: float prefactor for conversion to Steven's Coefficients
        """
        return self._alphalm.get((l, abs(m)), 0.0)

    def readPot(self, *args, lm=None, **kwargs):
        """Reads in the potentials for the CF coefficient calculation
        If hdf files are given also the muffin tin radius is read in

        Parameters:
            :param args: Expects string filenames for the potentials to read in
                         The function expects either HDF files or txt files with the
                         format (rmesh,vlmup,vlmdn)
            :param lm: list of tuples, Defines the l and m indices for the given txt files

        kwargs:
            :param atomType: int, Defines the atomType to read in (only for HDF files)
            :param header: int, Define how many lines to skip in the beginning of txt file
            :param complexData: bool, Define if the data in the text file is complex

        Raises:
            ValueError: lm indices list length has to match number of files read in

        """

        if lm is None:
            lm = []

        atomType = kwargs.get('atomType')
        header = kwargs.get('header', 0)
        complexData = kwargs.get('complexData', True)

        #Reads in the filenames given in args as potentials
        for index, file in enumerate(args):
            if isinstance(file, (str, Path)):
                basename, extension = os.path.splitext(file)
                if extension == '.hdf':
                    with h5py.File(file, 'r') as hdffile:
                        self.__readpotHDF(hdffile, atomType=atomType)
                else:
                    if index >= len(lm):
                        raise ValueError('Not enough lm indices for the given files')
                    self.__readpottxt(file, lm[index], header=header, complexData=complexData)
            else:
                self.__readpotHDF(file, atomType)

    def readCDN(self, file, **kwargs):
        """Reads in the normed charge density for the CF coefficient calculation
        If hdf files are given also the muffin tin radius is read in

        Parameters:
            :param file: Expects string filename for the charge density to read in
                         The function expects either HDF files or txt files with the
                         format (rmesh,cdn).
                         The charge density should be given as r^2n(r) and normed to 1
        kwargs:
            :param atomType: int, Defines the atomType to read in (only for HDF files)
            :param header: int, Define how many lines to skip in the beginning of txt file

        """

        atomType = kwargs.get('atomType')
        header = kwargs.get('header', 0)

        if isinstance(file, (str, Path)):
            basename, extension = os.path.splitext(file)
            if extension == '.hdf':
                with h5py.File(file, 'r') as hdffile:
                    self.__readcdnHDF(hdffile, atomType=atomType)
            else:
                self.__readcdntxt(file, header=header)
        else:
            self.__readcdnHDF(file, atomType)

    def __readpotHDF(self, hdffile, atomType=None):
        """Read in the potential from a HDF file

        """

        info = hdffile.get('general')
        numPOT = info.attrs['numPOT'][0]
        if 'bravaisMatrix' in info:
            self.bravaisMat['pot'] = np.array(info.get('bravaisMatrix'))

        if numPOT == 0:
            raise IOError(f'No potentials found in {hdffile}')

        potential_groups = {key for key in hdffile if 'pot-' in key}

        if len(potential_groups) != 1 and atomType is None:
            raise ValueError('Multiple possibilities for calculated potentials. '
                             f'Select the desired atomType: {potential_groups}')

        if atomType is not None:
            pot_group = f'pot-{atomType}'
        else:
            pot_group = potential_groups.pop()

        if pot_group in hdffile:
            _pot = hdffile.get(pot_group)

            self.vlm['RMT'] = _pot.attrs['RMT'][0]
            for key in _pot.keys():

                if key == 'rmesh':
                    _rmesh = _pot.get(key)
                    self.vlm['rmesh'] = np.array(_rmesh)
                else:
                    _vlm = _pot.get(key)
                    l = _vlm.attrs['l'][0]
                    m = _vlm.attrs['m'][0]

                    _data = _vlm.get('vlm')
                    _data = np.array(_data[:, :, 0] + 1j * _data[:, :, 1])
                    if abs(_data).max() >= self.pot_cutoff:
                        self.vlm[(l, m)] = _data

        else:
            raise IOError(f'No potential for atomType {atomType} found in {hdffile}')

        if not self.quiet:
            print(f'readPOTHDF: Generated the following information: {self.vlm.keys()}')

    def __readpottxt(self, file, index, header=0, complexData=True):
        """Read in the potential for the (l,m) tuple 'index' from a txt file
        The muffin-tin radius is infered from the biggest argument in the rmesh
        """

        with open(file, newline='') as csvfile:
            reader = csv.reader(csvfile, delimiter=' ', quotechar='|')

            if index not in self.vlm:
                self.vlm[index] = []
                self.vlm[index].append([])
                self.vlm[index].append([])
            else:
                raise KeyError(f'Multiple definitions for potential {index}')

            self.vlm['rmesh'] = []

            for line in skipHeader(reader, header):
                while '' in line:
                    line.remove('')
                self.vlm['rmesh'].append(float(line[0]))
                if complexData:
                    self.vlm[index][0].append(float(line[1]) + 1j * float(line[2]))
                    self.vlm[index][1].append(float(line[3]) + 1j * float(line[4]))
                else:
                    self.vlm[index][0].append(float(line[1]))
                    self.vlm[index][1].append(float(line[2]))

            self.vlm[index] = np.array(self.vlm[index])
            self.vlm['rmesh'] = np.array(self.vlm['rmesh'])
            self.vlm['RMT'] = max(self.vlm['rmesh'])

    def __readcdnHDF(self, hdffile, atomType=None):
        """Read in the charge density from a HDF file

        """

        info = hdffile.get('general')
        numCDN = info.attrs['numCDN'][0]
        if 'bravaisMatrix' in info:
            self.bravaisMat['cdn'] = np.array(info.get('bravaisMatrix'))

        if numCDN == 0:
            raise IOError(f'No charge densities found in {hdffile}')

        cdn_groups = {key for key in hdffile if 'cdn-' in key}

        if len(cdn_groups) != 1 and atomType is None:
            raise ValueError('Multiple possibilities for calculated charge densities. '
                             f'Select the desired atomType: {cdn_groups}')

        if atomType is not None:
            cdn_group = f'cdn-{atomType}'
        else:
            cdn_group = cdn_groups.pop()

        if cdn_group in hdffile:
            _cdn = hdffile.get(cdn_group)
            self.cdn['RMT'] = _cdn.attrs['RMT'][0]
            _rmesh = _cdn.get('rmesh')
            self.cdn['rmesh'] = np.array(_rmesh)
            _data = _cdn.get('cdn')
            self.cdn['data'] = np.array(_data)

        else:
            raise IOError(f'No charge density for atomType {atomType} found in {hdffile}')

        if not self.quiet:
            print(f'readcdnHDF: Generated the following information: {self.cdn.keys()}')

    def __readcdntxt(self, file, header=0):
        """Read in the charge density from a txt file
        The muffin-tin radius is infered from the biggest argument in the rmesh
        """

        with open(file, newline='') as csvfile:
            reader = csv.reader(csvfile, delimiter=' ', quotechar='|')

            if 'data' not in self.cdn:
                self.cdn['data'] = []
            else:
                raise KeyError('Charge density already read in')

            self.cdn['rmesh'] = []

            for line in skipHeader(reader, header):
                while '' in line:
                    line.remove('')

                self.cdn['rmesh'].append(float(line[0]))
                self.cdn['data'].append(float(line[1]))

            self.cdn['data'] = np.array(self.cdn['data'])
            self.cdn['rmesh'] = np.array(self.cdn['rmesh'])
            self.cdn['RMT'] = max(self.cdn['rmesh'])

    def _validateInput(self):
        """Validate that the object can be used to execute the calculation
        Checks that the given bravais matrices are equal if given
        """

        if 'data' not in self.cdn or 'rmesh' not in self.cdn:
            raise ValueError('Charge density input incomplete')

        if 'rmesh' not in self.vlm or len(self.vlm.items()) <= 2:
            raise ValueError('Potential input incomplete')

        if 'cdn' in self.bravaisMat and 'pot' in self.bravaisMat:
            diffBravais = self.bravaisMat['cdn'] - self.bravaisMat['pot']
            if np.any(np.abs(diffBravais) > 1e-8):
                raise ValueError('Differing definitions of potentials and charge density bravais matrix')

    def _interpolate(self):
        """Interpolate all quantities to a common equidistant radial mesh

        """

        if self.reference_radius == 'pot':
            refRMT = self.vlm['RMT']
        elif self.reference_radius == 'cdn':
            refRMT = self.cdn['RMT']
        else:
            refRMT = self.reference_radius

        self.int = defaultdict(dict)

        self.int['rmesh'] = np.arange(0.0, refRMT, refRMT / self.radial_points)
        self.int['cdn'] = interp1d(self.cdn['rmesh'], self.cdn['data'], fill_value='extrapolate')

        for key, value in self.vlm.items():
            if key not in ('RMT', 'rmesh'):
                self.int[key]['spin-up'] = interp1d(self.vlm['rmesh'], value[0, :], fill_value='extrapolate')
                if value.shape[0] == 2:
                    self.int[key]['spin-down'] = interp1d(self.vlm['rmesh'], value[1, :], fill_value='extrapolate')
        self.interpolated = True

    def performIntegration(self, convert=True):
        """Performs the integration to obtain the crystal field coefficients
        If the data was not already interpolated, the interpolation will
        be performed beforehand

        Parameters:
            :param convert: bool, converts to Steven's coefficients (if True)

        :returns: list of CFCoefficient objects (namedtuple), with all the necessary information

        """

        self._validateInput()

        if 'cdn' in self.bravaisMat and 'pot' in self.bravaisMat:
            a_vector = self.bravaisMat['pot'][0, :]
            c_vector = self.bravaisMat['pot'][2, :]

            self.theta = np.arccos(c_vector[2] / (np.linalg.norm(c_vector)))
            self.phi = np.arccos(a_vector[0] / (np.linalg.norm(a_vector)))

            if not self.quiet:
                print(fr'Angle between lattice vector c and z-axis: {self.theta/np.pi:5.3f} $\pi$')
                print(fr'Angle between lattice vector a and x-axis: {self.phi/np.pi:5.3f} $\pi$')

        if not self.interpolated:
            self._interpolate()

        self.denNorm = np.trapz(self.int['cdn'](self.int['rmesh']), self.int['rmesh'])
        if not self.quiet:
            print(f'Density normalization = {self.denNorm}')

        result = []
        for lmkey, vlm in [(key, val) for key, val in self.int.items() if isinstance(key, tuple)]:
            l, m = lmkey
            if not self.only_m0 or m == 0:
                integral = {}
                for key, pot in vlm.items():
                    integral[key] = np.trapz(
                        pot(self.int['rmesh']) * self.int['cdn'](self.int['rmesh']), self.int['rmesh'])
                    integral[key] *= np.sqrt((2.0 * l + 1.0) / (4.0 * np.pi)) * HTR_TO_KELVIN

                if 'spin-down' not in integral:
                    integral['spin-down'] = integral['spin-up']

                if convert:
                    integral = {key: val.real * self.prefactor(l, m) for key, val in integral.items()}
                    coeff = CFCoefficient(l=l,
                                          m=m,
                                          spin_up=integral['spin-up'],
                                          spin_down=integral['spin-down'],
                                          unit='K',
                                          convention='Stevens')
                else:
                    coeff = CFCoefficient(l=l,
                                          m=m,
                                          spin_up=integral['spin-up'],
                                          spin_down=integral['spin-down'],
                                          unit='K',
                                          convention='Wybourne')

                result.append(coeff)

        result.sort(key=lambda item: (item.l, abs(item.m)))

        if not self.quiet:

            print(f'\nThe following results were obtained with the {result[0].convention} convention:')

            if any(isinstance(coeff.spin_up, complex) for coeff in result):
                print('l  m', '       $C^{up}_{lm}$ [K]              ', '       $C^{dn}_{lm}$ [K]')
            else:
                print('l  m', '       $C^{up}_{lm}$ [K]', '       $C^{dn}_{lm}$ [K]')
            for coeff in result:
                if isinstance(coeff.spin_up, complex):
                    print(
                        f'{coeff.l:d}{coeff.m:>-3d}{coeff.spin_up.real:>+25.8f}{coeff.spin_up.imag:>+11.8f} i {coeff.spin_down.real:>+25.8f}{coeff.spin_down.imag:>+11.8f} i '
                    )
                else:
                    print(f'{coeff.l:d}{coeff.m:>-3d}{coeff.spin_up:>+25.8f}{coeff.spin_down:>+25.8f}')

        return result


def plot_crystal_field_calculation(cfcalc,
                                   filename='crystal_field_calc',
                                   pot_title='Potential',
                                   cdn_title='Density',
                                   xlabel='$R$ (Bohr)',
                                   pot_ylabel='$Vpot$ (Hartree)',
                                   cdn_ylabel='Density',
                                   fontsize=12,
                                   labelsize=12,
                                   pot_colors=None,
                                   save=False,
                                   show=True):
    """Plot the given potentials and charge densities

        Parameters:
            :param cfcalc: CFcalculation containing the data to plot
            :param filename: str, Define the filename to save the figure
            :param pot_title: Title for the potential subplot
            :param cdn_title: Title for the charge density subplot
            :param xlabel: label for the x axis of both subplots
            :param pot_ylabel: label for the y axis of the potential subplot
            :param cdn_ylabel: label for the y axis f the charge density subplot
            :param fontsize: fontsize for titles and labels on the axis
            :param labelsize: fontsize for the ticks on the axis,

        """

    cfcalc._validateInput()

    if not cfcalc.interpolated:
        cfcalc._interpolate()

    if pot_colors is None:
        pot_colors = ['black', 'red', 'blue', 'orange', 'green', 'purple']

    color_iter = iter(pot_colors)

    fig, axs = plt.subplots(1, 2)
    ax = axs[0]

    for lmkey, vlm in [(key, val) for key, val in cfcalc.int.items() if isinstance(key, tuple)]:
        l, m = lmkey
        if not cfcalc.only_m0 or m == 0:
            try:
                color = next(color_iter)
            except StopIteration:
                color = None
            line, = ax.plot(cfcalc.int['rmesh'],
                            vlm['spin-up'](cfcalc.int['rmesh']).real,
                            '-',
                            color=color,
                            label=rf'$V_{{{l}{m}}}$')
            ax.plot(cfcalc.int['rmesh'], vlm['spin-down'](cfcalc.int['rmesh']).real, '--', color=line.get_color())
    ax.set_xlabel(xlabel, fontsize=fontsize)
    ax.set_ylabel(pot_ylabel, fontsize=fontsize)
    ax.legend(loc=2, ncol=1, borderaxespad=0.0, fontsize=fontsize)
    ax.tick_params(labelsize=labelsize)
    ax.set_title(pot_title, fontsize=fontsize)

    ax = axs[1]
    ax.plot(cfcalc.cdn['rmesh'], cfcalc.cdn['data'], '-', color='black', label=r'$n(r)$')
    ax.plot(cfcalc.int['rmesh'], cfcalc.int['cdn'](cfcalc.int['rmesh']), '--', color='blue')
    ax.set_xlabel(xlabel, fontsize=fontsize)
    ax.set_ylabel(cdn_ylabel, fontsize=fontsize)
    ax.legend(loc=2, ncol=1, borderaxespad=0.0, fontsize=fontsize)
    ax.tick_params(labelsize=labelsize)
    ax.set_title(cdn_title, fontsize=fontsize)

    fig.set_size_inches(14.0, 10.0)
    fig.subplots_adjust(left=0.10, bottom=0.2, right=0.90, wspace=0.4, hspace=0.4)
    if save:
        plt.savefig(filename, format='png')
    if show:
        plt.show()


def plot_crystal_field_potential(cfcoeffs,
                                 filename='crystal_field_potential_areaplot',
                                 spin='avg',
                                 phi=0.0,
                                 save=False,
                                 show=True):
    """Plots the angular dependence of the calculated CF potential as well
        as a plane defined by phi.

        Parameters:
            :param cfcoeffs: list of CFCoefficients to construct the potential
            :param filename: str, defines the filename to save the figure
            :param spin: str; Either 'up', 'dn' or 'avg'. Which spin direction to plot
                         ('avg'-> ('up'+'dn')/2.0)
            :param phi: float, defines the phi angle of the plane

        :raises AssertionError: When coefficients are provided as wrong types or in the wrong convention

        """

    assert all(isinstance(coeff, CFCoefficient) for coeff in cfcoeffs), \
           'Only provide a list of CFCoefficients to plot_crystal_field_potential'

    #generate the thetha phi meshgrid
    theta_grid = np.linspace(0, np.pi, 181)
    phi_grid = np.linspace(0.0, 2.0 * np.pi, 361)
    xv, yv = np.meshgrid(phi_grid, theta_grid)
    cf_grid = np.zeros((181, 361), dtype='complex')

    for coeff in cfcoeffs:
        assert coeff.convention == 'Wybourne', 'Wrong convention for plotting in spherical harmonics basis'
        if spin == 'avg':
            value = 0.5 * (coeff.spin_up + coeff.spin_down)
        elif spin == 'up':
            value = coeff.spin_up
        elif spin == 'down':
            value = coeff.spin_down
        else:
            raise ValueError(f"Invalid Argument for spin: '{spin}'")
        if coeff.m >= 0:
            cf_grid += value * sph_harm(coeff.m, coeff.l, xv, yv)

    #Plot the angular dependence
    maxv = max(cf_grid.real.max(), abs(cf_grid.real.min()))
    tickFontsize = 14
    labelFontsize = 20
    #Angular dependence plot
    fig = plt.figure(figsize=(15, 5))
    gs = mpl.gridspec.GridSpec(1, 2, width_ratios=[2.0, 1.5])
    ax = plt.subplot(gs[0])
    plt.sca(ax)
    plt.imshow(cf_grid.real, origin='upper', cmap='coolwarm', vmin=-maxv, vmax=maxv, aspect='auto')
    ax.set_title('Angular Dependence', fontsize=labelFontsize)
    ax.set_xlabel(r'$\phi$', fontsize=labelFontsize)
    ax.set_xticks([0.0, 90.0, 180.0, 270.0, 360.0])
    ax.set_xticklabels([r'0', r'$\pi/2$', r'$\pi$', r'$3\pi/2$', r'$2\pi$'], fontsize=tickFontsize)
    ax.set_ylabel(r'$\theta$', fontsize=labelFontsize)
    ax.set_yticks([0.0, 45.0, 90.0, 135.0, 180.0])
    ax.set_yticklabels([r'$\pi/2$', r'$\pi/4$', r'$0.0$', r'$-\pi/4$', r'$-\pi/2$'], fontsize=tickFontsize)

    if np.abs(phi_grid - phi).min() > 1e-5:
        raise ValueError(f'Angle {phi} not found in grid')
    else:
        phi_ind = np.abs(phi_grid - phi).argmin()

    theta_grid_pm = list(-1.0 * theta_grid)
    theta_cf = list(cf_grid[:, phi_ind])
    for index, theta in enumerate(theta_grid):
        theta_grid_pm.append(theta)
        theta_cf.append(theta_cf[index])

    theta_grid = np.array(theta_grid)
    theta_cf = np.array(theta_cf)

    #interpolate
    theta_cf = interp1d(theta_grid_pm, theta_cf.real, fill_value='extrapolate')

    nx = 200
    ny = 200
    #Define the cartesian grid between -1 and 1
    x = np.linspace(-1, 1, nx)
    y = np.linspace(-1, 1, ny)
    xv, yv = np.meshgrid(x, y)

    z = theta_cf(np.arctan(xv / yv))

    phi_fract = phi / np.pi

    labelFontsize = 14
    tickFontsize = 14

    ax = plt.subplot(gs[1])
    plt.sca(ax)
    plt.imshow(z, origin='upper', cmap='coolwarm', vmin=-maxv, vmax=maxv, aspect='auto')
    cbar = plt.colorbar(shrink=0.8)
    cbar.set_label(r'$V_{{CF}}$ [K]', fontsize=labelFontsize)
    cbar.ax.tick_params(labelsize=14)
    ax.set_title(r'Crystal Field potential for $\phi={{{:.2f}}}\pi$'.format(phi_fract), fontsize=labelFontsize)
    ax.set_xlabel(r'x [Bohr]', fontsize=labelFontsize)
    ax.set_xticks([0, nx / 4.0, nx / 2.0, 3.0 * nx / 4.0, nx - 1])
    ax.set_xticklabels([r'-1.0', r'-0.5', r'0.0', r'0.5', r'1.0'], fontsize=tickFontsize)
    ax.set_ylabel(r'y [Bohr]', fontsize=labelFontsize)
    ax.set_yticks([0, ny / 4.0, ny / 2.0, 3.0 * ny / 4.0, ny - 1])
    ax.set_yticklabels([r'1.0', r'0.5', r'0.0', r'-0.5', r'-1.0'], fontsize=tickFontsize)
    fig.set_constrained_layout_pads(w_pad=0., h_pad=0.0, hspace=0., wspace=0.)

    if save:
        plt.savefig(filename, format='png')
    if show:
        plt.show()
